#ifndef AMATRIX_H
#define AMATRIX_H

#include <rbtree.h>

#define MAXNDIMS 10

typedef int SubmatId;
typedef int RowId;
typedef int TableId;
typedef int ValueId;

struct amatrix {
    int ndims;
    int eltsize;
    TableId root;
    struct tstore *submats;
    struct aarray *entries;
    struct aarray *values;
};    

extern struct amatrix *AMatrix;

int submat_length (int i, struct amatrix *m);
int submat_nonempty_components (int i, struct amatrix *m);
void *am_insert (int *key, struct amatrix *m);
void *am_search (int *keytuple, struct amatrix *m);
SubmatId submat_id (int *key, int n);

typedef struct tree_state ComponentList;
    
struct am_row {
    SubmatId i;
    int keytuple[MAXNDIMS];
    struct amatrix *m;
};

struct am_cell {
    int key;
    void *value;
};

struct am_cells {
    struct am_row *row;
    ComponentList components;
};

int am_get_row (struct am_row *row, int *keytuple, struct amatrix *m);
int am_start_row (struct am_cells *iter, struct am_row *row);
int am_next_cell (struct am_cell *cell, struct am_cells *iter);

struct am_rows {
    struct amatrix *mat;
    int keytuple[MAXNDIMS];
    int n;
    int d0;
    int dN;
    ComponentList components[MAXNDIMS];
};

int am_start_submats (struct am_rows *iter, int targdim, int *keytuple, int n,
		      struct amatrix *m);
SubmatId am_next_submat (struct am_rows *iter);
SubmatId am_next_bottom_submat (struct am_rows *iter);

#define am_start_rows(iter,m)  (am_start_submats((iter), 1, NULL, 0, (m)))

int am_next_row (struct am_row *row, struct am_rows *iter);

struct amatrix *make_amatrix (int ndims, int alloc, int eltsize);
int write_amatrix (struct amatrix *m, FILE *file);
struct amatrix *read_amatrix (FILE *file);
int dump_amatrix (struct amatrix *m, int (*print_value)(), FILE *file);
int print_amatrix (struct amatrix *m, int (*print_elt)(), FILE *file);


#endif
