#include <stdio.h>
#include <string.h>
#include <spa.h>
#include <amgr.h>
#include <lists.h>
#include <dict.h>
#include <strtab.h>
#include <hash.h>
#include <pool.h>
#include <reg.h>
#include <casstrace.h>


#define SENTLEN      128  /* Sentences exceeding this length are split */
#define MAXINPUTLINE 128

struct string_array *Words = NULL;
struct phrase *Chart0[SENTLEN];
struct phrase **Chart = Chart0-1;
char NewPhrases[(SENTLEN+1)/8];
struct pool *Pool = NULL;
int CorpusPosition = -1;

#define corpus_position(i) (CorpusPosition + (i))
#define print_tree(p,file) (print_tree_1(p,NULL,file), terpri(file))

FILE *Outfile;
FILE **Outfiles;
int (*print_level)() = NULL;
int (*print_sentence)() = NULL;
int PreserveSgml = 0;
int PrintChildren = 1;
int MultiFile = 0;
int DontPrintAttributes = 1;
int DontPrintRoles = 1;
int DontPrintCPs = 1;
SymbolId CassGrepCat = 0;

int MatchVerbose = 0;
int NoActions = 0;
int NoWarn = 0;
char *Bos = "<s>";
char *Eos = "</s>";
int BosSeen = 0;
int EosPending = 0;
char EosTag[MAXINPUTLINE];
char *Head = "h";

int TmpMATCHi;
struct phrase *TmpMATCHp;

union value {
    t_subset x;
    struct phrase *p;
    char *s;
};

union value *DefaultVA;
union value *TmpVA;

struct phrase {
    int cat;
    int lvl;
    int i;
    int j;
    union value *va;           /* Features.  Length = NAtts. */
    struct phrase **children;  /* Terminated by NULL */
};

#define att_xval(ph,f) ((ph)->va[(f)-1].x)
#define att_pval(ph,f) ((ph)->va[(f)-1].p)
#define att_sval(ph,f) ((ph)->va[(f)-1].s)

#define phrase_child(ph,att) ((ph)->va == NULL ? NULL : (ph)->va[(att)-1].p)

#define INTERMEDIATE 1
    
#define make_phrase(CAT,I,J) \
  (TmpMATCHp = palloc(1, sizeof(struct phrase), Pool), \
   TmpMATCHp->cat = (CAT), \
   TmpMATCHp->lvl = CurrentLevel, \
   TmpMATCHp->i = (I), \
   TmpMATCHp->j = (J), \
   TmpMATCHp)


#define make_value_array() \
  (TmpVA = palloc(NAtts, sizeof(union value), Pool), \
   bcopy(DefaultVA, TmpVA, NAtts*sizeof(union value)), \
   TmpVA)

#define make_children(n)   (palloc((n)+1, sizeof(struct phrase *), Pool))

#define SentLen (string_array_len(Words))

#define word_string(i) (string_array_ref(Words, ((i)-1)))
#define insert_word(s) (string_array_insert((s), Words))


print_phrase (struct phrase *p, FILE *file)
{
    struct phrase **children = p->children;
    
    if (p == NULL) fprintf(file, "NULL");
    else {
	fprintf(file, "<%s %d", p->cat ? symbol_string(p->cat) : "no cat", p->i);
	if (children == NULL) {
	    fprintf(file, " %d", p->j);
	}
	else {
	    for (; *children != NULL; children++) {
		fprintf(file, " %s %d", symbol_string(children[0]->cat), children[0]->j);
	    }
	}
	fprintf(file, ">");
    }
}


struct phrase *phrase_child_rec (struct phrase *p, AttId role)
{
    struct phrase *n;

    while ((n = phrase_child(p, role)) != NULL) {
	p = n;
    }
    return p;
}


struct psnode {
    int s;
    int i;
    int local;
};

#define STACKSIZE 128

int *find_elocal_positions (int cat, int start, int end);

init_match ()
{
    int i;
    AttId a;

    if (Words == NULL) {
	Words = make_string_array(SENTLEN);
	Pool = make_pool(SENTLEN * (NLevels+1) * sizeof(int));
	allocate(DefaultVA, NAtts, sizeof(union value));
	for (a = 1; a <= NAtts; a++) {
	    if (att_type(a) == ATT_SUBSET) {
		DefaultVA[a-1].x = ~(t_subset)0;
	    }
	}
    }
    else {
	CorpusPosition += SentLen;
	zero_string_array(Words);
	zero_pool(Pool);
    }
}



/**===================================================================
***
***			       PHRASES
**/


#define print_chart(file) (print_chart(1,SentLen+1,(file)))


print_subchart (int start, int end, FILE *file)
{
    int i;

    print_char('[', file);
    for (i = start; i < end; i = Chart[i]->j) {
	if (i > 1) print_char(' ', file);
	fprintf(file, "%d %s", i, Chart[i]->cat ? symbol_string(Chart[i]->cat) : "(no cat)");
    }
    fprintf(file, " %d", i);
    print_char(']', file);
}


dump_chart (FILE *file)
{
    int i, j, k;
    struct phrase *p;
    
    fprintf(file, "Chart:\n");
    push_indent(2);
    for (i = j = 1; i <= SentLen; i++) {
	indent(file);
	fprintf(file, "%2d ", i);
	print_field(12, word_string(i), file);
	if (i >= j && (p = Chart[i]) != NULL) {
	    fprintf(file, " %s", p->cat ? symbol_string(p->cat) : "(no cat)");
	    if (p->va != NULL) {
		for (k = 1; k <= NAtts; k++)
		  fprintf(file, " %d", att_xval(p,k));
	    }
	    if (p->j <= i) error("dump_chart: j <= i");
	    fprintf(file, " j=%d", p->j);
	    j = p->j;
	}
	fprintf(file, "\n");
    }
    pop_indent();
}



/**===================================================================
***
***			    SCAN SENTENCE
***
***  Input file format: "tag word\n", where
***  tags are strings, not indices.  Needs a word_array and
***  the vocabulary as well as a phrase_array.
**/

/**
***  The vocabulary initially contains only those tags that are explicitly
***  mentioned in the grammar.  Additional tags are added as sentences are
***  scanned.  I don't think I'm making any dangerous assumptions based on
***  a fixed vocabulary size.
**/

int scan_sentence (FILE *file)
{
    char c;
    int x, word, cat, i, delim;
    struct record r;
    struct phrase *ph;

    skip_whitespace(file);
    if (peek_char(file) == EOF) return EOF;

    init_match();
    use_level(0);
    BosSeen = 0;
    for (i = 1; peek_char(file) != EOF; i++) {
	if (i > SENTLEN) {
	    if (!NoWarn)
	      fprintf(stderr, "Run-on sentence at %d, splitting\n", ftell(file));
	    break;
	}
	if (scan_record(&r, file) == EOF) break;
	if (r.nfields == 1) {
	    if (i == 1) {
		if (PreserveSgml) {
		    if (!DontPrintCPs) print_string("          ", Outfile);
		    fprintf(Outfile, "%s\n", r.fields[0]);
		}
		i--;  /* to cancel the increment */
	    }
	    else {
		EosPending = 1;
		strcpy(EosTag, r.fields[0]);
		break;
	    }
	}
	else {
	    word = insert_word(r.fields[0]);
	    cat = intern(r.fields[1]);
	    Chart[i] = ph = make_phrase(cat, i, i+1);

	    if (!NoActions) {
		int k, n;
		char *s, **fields = &r.fields[0] - 1;
		union value *va = ph->va = make_value_array();
		AttId a;

		for (a = 1; a <= NAtts; a++) {
		    if ((k = att_field(a)) && r.nfields >= k) {
			switch (att_type(a)) {
			  case ATT_STRING:
			    n = strlen(fields[k]);
			    s = palloc(n+1, sizeof(char), Pool);
			    strcpy(s, fields[k]);
			    va[a-1].s = s;
			    break;
			  case ATT_PHRASE:
			    error("phrase att with field");
			  case ATT_SUBSET:
			    va[a-1].x = atosubset(fields[k], att_range(a));
			    break;
			}
		    }
		}
	    }
	}
    }
    return i;
}


inline print_eos (FILE *file)
{
    if (EosPending) {
	if (!DontPrintCPs) print_string("          ", file);
	print_string(EosTag, file);
	terpri(file);
	EosPending = 0;
    }
}    


/**===================================================================
***
***			  Match, Match Level
**/
    

match ()
{
    int i;
    for (i = 1; i <= NLevels; i++) {
	match_level(i);
	if (print_level != NULL) print_level();
    }
}


/**  Returns non-zero if there was a hit at this level. **/

int match_level (int i)
{
    int start = 1, s, local, *finals, cmax, jmax, hit = 0;
    struct phrase *in, *out;

    use_level(i);
    use_fsa(level_fsr(CurrentLevel));
    zero_bit_array(NewPhrases, SentLen+1);  /* the +1 is because we don't use position 0 */
    Trace_M_start;

    while (start <= SentLen) {
	i = start; jmax = -1; s = start_state;
	Trace_M_loop;

	while (s && i <= SentLen) {
	    in = Chart[i];

	    if (in->cat == NullCat && i > start) {
		Trace_M_nil;
		i = in->j;
		continue;
	    }

	    local = cat_local(in->cat);
	    if (!local) {
		Trace_M_unrecognized;
		break;
	    }

	    s = state_next1(s, local);
	    i = in->j;
	    Trace_M_next;
	    if (s && (finals = state_finals(s))) {
		cmax = *finals;
		jmax = i;
		Trace_M_finals;
	    }
	}

	/* Successful match */
	if (jmax != -1) {
	    Trace_M_match;
	    if (cmax == CassGrepCat) grep_print(cmax, start, jmax);
	    start = parse_phrase(cmax, start, jmax);
	    hit = 1;
	}

	/* Punt */
	else {
	    in = Chart[start];
	    Trace_M_punt;
	    start = in->j;
	}
    }
    Trace_M_done;

    return hit;
}



/**===================================================================
***
***			     Parse Phrase
**/

struct phrase *set_children (struct phrase *p)
{
    int i, k, nkids = 0;

    for (i = p->i; i < p->j; i = Chart[i]->j) nkids++;
    p->children = make_children(nkids);
    for (i = p->i, k = 0; i < p->j; i = Chart[i]->j, k++) {
	p->children[k] = Chart[i];
	/* Intermediate nodes */
	if (Chart[i]->lvl == CurrentLevel) Chart[i]->lvl = -1;
    }
    return p;
}	

/**************************************************************************/
/**                                                                      **/
/**  Find positions for action locals.  Fills an array iv so that `k =   **/
/**  iv[local-ILocal]' is the position in the input where action local   **/
/**  `local' should be executed.  `buf+galloc(..)' is a zero-based       **/
/**  array.  We're addressing it by local-ILocals, which begins at 1.    **/
/**  Hence the extra `- 1' where iv is initialized.                      **/
/**                                                                      **/
/**  The stack contains state info (configurations) for backtracking.    **/
/**  A configuration includes the current fst state, the current input   **/
/**  position, and the next action local to try.                         **/
/**                                                                      **/
/**  An action local of -1 means that we have not yet taken a            **/
/**  transition from the configuration.  We update the -1 to the value   **/
/**  of ILocals (one less than the first action local) and compute the   **/
/**  transition.  If there is a transition, we push a new                **/
/**  configuration on the stack.                                         **/
/**                                                                      **/
/**  As long as there are transitions, we keep pushing configurations    **/
/**  on the stack.  If we reach the end in a final state, we're done.    **/
/**  If we run out of next states before reaching the end, we end up     **/
/**  with a 'used' configuration (local != -1) on top of the stack.      **/
/**                                                                      **/
/**  If the state on top of the stack is not fresh, we increment its     **/
/**  local and try a transition on the result, which is an action        **/
/**  local.  If there is a transition on that local, we push it on the   **/
/**  stack, otherwise we increment the local and try again until we      **/
/**  find one with a transition.  If there are no transitions on any     **/
/**  action locals, we pop the stack.                                    **/
/**                                                                      **/
/**  Finally we read the sequence of action locals off the stack.        **/
/**  Each configuration on the stack whose local is greater than         **/
/**  ILocals represents an action transition.  It is possible that one   **/
/**  and the same action transition will occur at multiple points in     **/
/**  the input; in that case we take the rightmost position.  That is,   **/
/**  we pass through the stack from left to right, setting               **/
/**  `iv[local-ILocals]' to `k', for `k' the most recent position at     **/
/**  which action local `local' was used.                                **/
/**                                                                      **/
/**  We could modify this to handle multiple-assignment, but the         **/
/**  issues aren't clear to me, so I'm leaving it this way for now.      **/
/**                                                                      **/
/**************************************************************************/


int *find_elocal_positions (int cat, int start, int end)
{
    static struct amgr *buf_mgr;
    static int *buf;
    int *iv, top, i, next, *finals;
    struct psnode stack[STACKSIZE];

    if (buf == NULL) buf_mgr = make_amgr(&buf, 50, sizeof(int));
    else gzero(buf_mgr);
    iv = buf + galloc(buf_mgr, NLocals - ILocals) - 1;
    
    stack[0].s = start_state;
    stack[0].i = start;
    stack[0].local = -1;
    top = 0;

    Trace_FEP_start;

    while (1) {
	i = stack[top].i;
	if (i == end
	    && (finals = state_finals(stack[top].s)) != NULL 
	    && *finals == cat) {
	    Trace_FEP_success;
	    break;
	}

	else if (stack[top].local < 0) {
	    stack[top].local = ILocals;
	    if (i < end && (next = Chart[i]->cat == NullCat ? stack[top].s :
			    state_next(stack[top].s, Chart[i]->cat))) {
		if (++top >= STACKSIZE) error("parse_phrase: stack overflow");
		stack[top].s = next;
		stack[top].i = Chart[i]->j;
		stack[top].local = -1;
	    }
	}
	else {
	    Trace_FEP_backtrack;
	    next = 0;
	    while (!next && ++stack[top].local <= NLocals) {
		next = state_next1(stack[top].s, stack[top].local);
	    }
	    if (next) {
		if (++top >= STACKSIZE) error("parse_phrase: stack overflow");
		stack[top].s = next;
		stack[top].i = i;
		stack[top].local = -1;
	    }
	    else if (--top < 0) error("parse_phrase: no parse");
	}
    }
    
    /** Now read elocal positions off the stack **/
    for (i = 0; i <= top; i++) {
	if (stack[i].local > ILocals)
	  iv[stack[i].local-ILocals] = stack[i].i;
    }
    Trace_FEP_done;
    return iv - ILocals;
}


int parse_phrase (int cat, int start, int end)
{
    struct phrase *p = NULL;
    FsaId fsa, prev_fsa;

    Trace_P_start;

    if (NoActions || !(fsa = cat_fst(cat))) {
	Trace_P_noparse;
    }

    else {
	struct phrase *c = NULL, *child;
	int e, f, i, m, *iv;
	t_subset sub;

	prev_fsa = CurrentFsa;
	use_fsa(fsa);

	iv = find_elocal_positions(cat, start, end);
	Trace_P_iv;
	for (e = ILocals + 1; e <= NLocals; e++) {
	    if (iv[e]) {
		switch (elocal_type(e)) {

		  case ACT_RIGHT:
		    Trace_P_right;
		    c = make_phrase(local_cat(e), 0, iv[e]);
		    break;

		  case ACT_LEFT:
		    c->i = iv[e];
		    Chart[iv[e]] = set_children(c);
		    setbit(NewPhrases, iv[e]);
		    Trace_P_left;
		    c = p;
		    break;

		  case ACT_ASSIGN:
		    if (c == NULL) {
			c = p = set_children(make_phrase(cat, start, end));
		    }
		    if (c->va == NULL) {
			c->va = make_value_array();
		    }
			
		    f = symbol_att(local_cat(e));
		    child = Chart[iv[e]];
		    Trace_P_assign;
		    switch (att_type(f)) {
		      case ATT_STRING:
			/* only legal mode is ATT_SET */
			if (child->va != NULL) {
			    att_sval(c,f) = att_sval(child,f);
			}
			break;
		      case ATT_PHRASE:
			/* only legal mode is ATT_SET */
			att_pval(c,f) = child;
			break;
		      case ATT_SUBSET:
			if (child->va != NULL) {
			    switch (att_mode(f)) {
			      case ATT_AND:
				att_xval(c,f) &= att_xval(child,f);
				break;
			      case ATT_IOR:
				att_xval(c,f) |= att_xval(child,f);
				break;
			      case ATT_XOR:
				att_xval(c,f) ^= att_xval(child,f);
				break;
			      case ATT_SET:
				att_xval(c,f) = att_xval(child,f);
				break;
			      default:
				error("bad att mode: %d", att_mode(f));
			    }
			}
			break;
		      default:
			error("bad att type: %d", att_type(f));
		    }
		}
	    }
	}
	use_fsa(prev_fsa);
    }

    if (p == NULL) p = set_children(make_phrase(cat, start, end));
    Chart[start] = p;
    setbit(NewPhrases, start);
    Trace_P_end;
    return end;
}



/**===================================================================
***
***			      Grep Print
**/

grep_print (SymbolId cat, int start, int end)
{
    int i;

    terpri(stdout);
    for (i = 1; i <= SentLen; i = Chart[i]->j) {
	if (i == start) {
	    printf("*********************************<%s>\n", symbol_string(cat));
	}
	if (i == end) {
	    printf("*********************************</%s>\n", symbol_string(cat));
	}
	print_tree(Chart[i], Outfile);
    }
    print_eos(Outfile);
}


/**===================================================================
***
***			    Forest Format
**/


#define forest_format() (print_sentence = &print_forest, PreserveSgml = 1)


int child_role (struct phrase *child, struct phrase *parent)
{
    if (parent->va != NULL) {
	int f;
	for (f = 1; f <= NAtts; f++) {
	    if (att_type(f) == ATT_PHRASE
		&& att_pval(parent,f) == child) return att_symbol(f);
	}
    }
    return 0;
}


print_tree_1 (struct phrase *p, struct phrase *par, FILE *file)
{
    int k, r;
    struct phrase **c;

    if (!DontPrintCPs) fprintf(file, "%08d", corpus_position(p->i));
    indent(file);
    if (!DontPrintRoles && par != NULL && (r = child_role(p, par))) {
	fprintf(file, "%c=", symbol_string(r)[0]);
    }
    else fprintf(file, "  ");
    push_indent(2);

    fprintf(file, "[%s", symbol_string(p->cat));
    print_features(p, file);
    c = p->children;
    if (c == NULL) {
	for (k = p->i; k < p->j; k++) {
	    fprintf(file, " %s", word_string(k));
	}
	fprintf(file, "]");
    }
    else {
	terpri(file);
	while (*c != NULL) {
	    print_tree_1(*c, p, file);
	    c++;
	    if (*c != NULL) terpri(file);
	}
	fprintf(file, "]");
    }
    pop_indent();
}


print_features (struct phrase *p, FILE *file)
{
    int i, mode;
    struct phrase *v;
    char *s;

    if (!DontPrintAttributes && p->va != NULL) {
	for (i = 1; i <= NAtts; i++) {
	    switch (att_type(i)) {
	      case ATT_STRING:
		s = att_sval(p,i);
		fprintf(file, " %s=%s", att_string(i), s==NULL?"":s);
		break;
	      case ATT_PHRASE:
/*
 *		v = att_pval(p,i);
 *		if (v == NULL)
 *		  fprintf(file, " %s=(null)", att_string(i));
 *		else
 *		    fprintf(file, " %s=<%s %d>",
 *			    att_string(i), symbol_string(v->cat), corpus_position(v->i));
 */
		break;
	      case ATT_SUBSET:
		fprintf(file, " %s=", att_string(i));
		print_subset(att_xval(p,i), att_range(i), file);
		break;
	    }
	}
    }
}

int sprint_features (struct phrase *p, char *s, int alloc)
{
    int i, off=0;
    char *t;
    
    if (!DontPrintAttributes && p->va != NULL) {
	for (i = 1; i <= NAtts; i++) {
	    sprintf(s+off, " %s=", att_string(i));
	    off += strlen(s+off);
	    switch (att_type(i)) {
	      case ATT_STRING:
		t = att_sval(p,i);
		if (t == NULL) sprintf(s+off, "none");
		else sprintf(s+off, "\"%s\"", t);
		off += strlen(s+off);
		break;
	      case ATT_PHRASE:
/*
 *		sprintf(s+off, "0x%x", att_pval(p,i));
 *		off += strlen(s+off);
 */
		break;
	      case ATT_SUBSET:
		sprint_subset(att_xval(p,i), att_range(i), s+off, alloc-off);
		off += strlen(s+off);
	    }
	}
    }
    return off;
}

print_forest ()
{
    int i;

    for (i = 1; i <= SentLen; i = Chart[i]->j) {
	print_tree(Chart[i], Outfile);
    }
    print_eos(Outfile);
}


/**===================================================================
***
***		      Tuple Format, Pair Format
**/

int HeadFtr = 0;
int C, C0, NX, NP, NG, NMESS, PP;
int print_tuples();
int PairStyle = 0;

char Tab1 = '\t';
char Tab2 = '\t';

tuple_format ()
{
    if (!HeadFtr) error("init_tuple_format: no head feature");
    C = intern("c");
    C0 = intern("c0");
    NX = intern("nx");
    NP = intern("np");
    NG = intern("ng");
    NMESS = intern("nmess");
    PP = intern("pp");
    print_sentence = &print_tuples;
}

#define pair_format() (tuple_format(), PairStyle = 1)
#define role_format() (tuple_format(), PairStyle = 1, Tab1 = ':')
#define arg_format()  (tuple_format(), PairStyle = 1, Tab2 = ':')

char *TmpTUPs;
#define phrase_headword(p) \
  (p->va == NULL ? NULL : att_sval(p,HeadFtr))

print_tuples ()
{
    char *head;
    int i;
    
    for (i = 1; i <= SentLen; i = Chart[i]->j) {
	if (Chart[i]->cat == C || Chart[i]->cat == C0) {
	    head = phrase_headword(Chart[i]);
	    if (PairStyle) print_tuple(Chart[i], head, corpus_position(i));
	    else {
		fprintf(Outfile, "%d\t%s", corpus_position(i), head == NULL ? ":none" : head);
		print_tuple(Chart[i], head, 0);
		print_char('\n', Outfile);
	    }
	}
    }
}

print_tuple (struct phrase *p, char *head, int cpos)
{
    char *h, *s;
    int cat;
    struct phrase *c, *hc, **children = p->children;

    if (children != NULL) {
	for (c = *children++; c != NULL; c = *children++) {
	    if (c->cat == C || c->cat == C0) {
		print_tuple(c, head, cpos);
	    }
	    else {
		h = phrase_headword(c);
		if (head == NULL || h != head) {
		    if (c->cat == NX || c->cat == NP || c->cat == NG || c->cat == NMESS) {
			if (p->cat == C0) s = "subj";
			else s = "obj";
		    }
		    else if (c->cat == PP) {
			hc = c->children[0];
			s = hc->va == NULL ? NULL : att_sval(hc, HeadFtr);
			if (s == NULL) s = "prep";
		    }
		    else s = symbol_string(c->cat);
		    if (h == NULL) h = ":no-head";
		    if (PairStyle) {
			if (!DontPrintCPs) fprintf(Outfile, "%d\t", cpos);
			print_string(head == NULL ? ":none" : head, Outfile);
			print_char(Tab1, Outfile);
			print_string(s, Outfile);
			print_char(Tab2, Outfile);
			print_string(h, Outfile);
			terpri(Outfile);
		    }
		    else fprintf(Outfile, "\t:%s %s", s, h);
		}
	    }
	}
    }
}



/**===================================================================
***
***			   Vertical Format
**/


#define vertical_format() (print_sentence = &print_vertical, PreserveSgml = 1)


#define level_indent(i, lvl, file) \
  for (i = NLevels; i >= lvl; i--) print_string("  ", file)


print_tree_vertical (int r, struct phrase *p, FILE *file)
{
    int i, k;
    struct phrase **c;

    if (p->lvl == 0) {
	level_indent(i, 0, file);
	print_preterm(r, p, 30, file);
	if (!DontPrintCPs) fprintf(file, "%07d ", corpus_position(p->i));
	fprintf(file, "%s\n", word_string(p->i));
    }
    else {
	if (p->lvl > 0) {
	    level_indent(i, p->lvl, file);
	    fprintf(file, "[%s", symbol_string(p->cat));
	    if (r && !DontPrintRoles) fprintf(file, ":%s", symbol_string(r));
	    print_features(p, file);
	    if (p->lvl == 0) print_char(']', file);
	    terpri(file);
	}
	c = p->children;
	if (c == NULL) {
	    for (k = p->i; k < p->j; k++) {
		level_indent(i, -1, file);
		if (!DontPrintCPs) fprintf(file, "%07d ", corpus_position(k));
		fprintf(file, "%s\n", word_string(k));
	    }
	}
	else {
	    for (; *c != NULL; c++) {
		print_tree_vertical(child_role(*c, p), *c, file);
	    }
	}
	level_indent(i, p->lvl, file);
	fprintf(file, "]\n");
    }
}


print_preterm (int role, struct phrase *p, int width, FILE *file)
{
    int i, len, mode;
    char buf[256], *s = &buf[0];

    sprintf(s, "%s", symbol_string(p->cat));
    s += strlen(s);

    if (role && !DontPrintRoles) {
	sprintf(s, ":%s", symbol_string(role));
	s += strlen(s);
    }

    s += sprint_features(p, s, 256-(s-buf));

    for (len = s - buf; len < width; len++) {
	*s++ = ' ';
    }
    buf[width-1] = ' ';
    buf[width] = 0;
    print_string(buf, file);
}


print_vertical ()
{
    int i;

    for (i = 1; i <= SentLen; i = Chart[i]->j) {
	print_tree_vertical(0, Chart[i], Outfile);
    }
    print_eos(Outfile);
}
    

/**===================================================================
***
***			    Binary Format
**/

#define binary_format() (MultiFile = 1, print_level = &write_multi)

int NodeCount = 0;

write_multi ()
{
    int i;
    FILE *file = Outfiles[CurrentLevel];
    
    for (i = 1; i <= SentLen; i++) {
	if (getbit(NewPhrases, i)) {
	    write_int(Chart[i]->cat, file);
	    write_int(Chart[i]->j, file);
	}
	else {
	    write_int(0, file);
	    write_int(0, file);
	}
    }
}


/**===================================================================
***
***			     Multi Format
**/


#define multi_format() (MultiFile = 1, print_level = &print_multi)

print_multi ()
{
    int i;
    FILE *file = Outfiles[CurrentLevel];
    
    for (i = 1; i <= SentLen; i++) {
	if (getbit(NewPhrases, i)) {
	    fprintf(file, "%s %d\n", symbol_string(Chart[i]->cat), Chart[i]->j - i);
	}
	else {
	    terpri(file);
	}
    }
}


/**===================================================================
***
***			    Phrases Format
**/

#define phrases_format() (print_level = &print_phrases)

print_phrases ()
{
    int i;
    char *s;
    struct phrase *p, *child, **children;
    
    for (i = 1; i <= SentLen; i++) {
	if (getbit(NewPhrases, i)) {
	    p = Chart[i];

	    fprintf(Outfile, "%s\t%d\t", symbol_string(p->cat), corpus_position(i));
	    if (PrintChildren && p->children != NULL) {
		for (children = p->children; *children != NULL; children++) {
		    child = children[0];
		    print_char(' ', Outfile);
		    print_string(symbol_string(child->cat), Outfile);
		    if (HeadFtr && child->va != NULL && (s = att_sval(child, HeadFtr))) {
			print_char(':', Outfile);
			print_string(s, Outfile);
		    }
		}
	    }
	    else {
		print_char(' ', Outfile);
		print_words(p->i, p->j, Outfile);
	    }
	    terpri(Outfile);
	}
    }
}


print_words (int i, int j, FILE *file)
{
    int k;
    
    for (k = i; k < j; k++) {
	if (k > i) print_char(' ', file);
	print_string(word_string(k), file);
    }
}


/**===================================================================
***
***			     Xkwic Format
**/

#define xkwic_format() (MultiFile = 1, print_level = &xkwic_print_multi)

xkwic_print_multi ()
{
    int i, k;
    char *cat;
    FILE *file = Outfiles[CurrentLevel];
    struct phrase *p;
    
    for (i = 1; i <= SentLen; i = p->j) {
	p = Chart[i];
	if (getbit(NewPhrases, i)) {
	    cat = symbol_string(p->cat);
	    for (k = i; k < p->j; k++) {
		print_char(k == i ? '<' : '-', file);
		print_char(k == p->j-1 ? '>' : '-', file);
		print_string(cat, file);
		terpri(file);
	    }
	}
	else {
	    for (k = i; k < p->j; k++) {
		print_string("--\n", file);
	    }
	}
    }
}


/**===================================================================
***
***				 Main
**/


init_outfiles (char *filename)
{
    int i;
    char *suf, buffer[MAXFILENAME];
    
    if (filename == NULL) error("init_outfiles: no filename");

    allocate(Outfiles, NLevels, FILE *);
    Outfiles--;    /* so that it is indexed from 1 */

    if (string_equal(filename, "/dev/null")) {
	for (i = 1; i <= NLevels; i++) {
	    Outfiles[i] = Fopen(filename, "w");
	}
    }
    else {
	strcpy(buffer, filename);
	filename = buffer;
	chsuf(filename, ".tag", ".");
    
	for (i = 1; i <= NLevels; i++) {
	    suf = symbol_string(level_name(i));
	    addsuf(filename, suf);
	    Outfiles[i] = Fopen(filename, "w");
	    fprintf(stderr, "Writing %s\n", filename);
	    rmsuf(filename, suf);
	}
    }
}


load_cass_grammar (char *grammar)
{
    if (grammar == NULL) {
	defaults(grammar, "CASS_GRAMMAR");
	if (grammar == NULL) grammar = DEFAULT_CASS_GRAMMAR;
    }
    if (has_suf(grammar, "reg")) load_reg(grammar);
    else load_fsc(grammar);
}
