#ifndef CASS_H
#define CASS_H

#include <reg.h>
#include <regdict.h>
#include <strtab.h>

#define SENTLEN    128  /* Sentences exceeding this length are split */
#define BUFFERSIZE 128

extern struct string_array *Words;
extern struct phrase *Chart0[SENTLEN];
extern struct phrase **Chart;
extern char NewPhrases[(SENTLEN+1)/8];
extern struct pool *Pool;
extern int CorpusPosition;

#define corpus_position(i) (CorpusPosition + (i))

extern FILE *Outfile;
extern FILE **Outfiles;
extern int (*print_level)();
extern int (*print_sentence)();
extern int PreserveSgml;
extern int PrintChildren;
extern int MultiFile;
extern int NoCpos;
extern int DontPrintAttributes;
extern int DontPrintRoles;
extern int DontPrintCPs;
extern SymbolId CassGrepCat;

extern int MatchVerbose;
extern int NoActions;
extern int NoWarn;
extern char *Bos;
extern char *Eos;
extern char *Head;

union value {
    t_subset x;
    struct phrase *p;
    char *s;
};

struct phrase {
    int cat;
    int lvl;
    int i;
    int j;
    union value *va;           /* Features.  Length = NAtts. */
    struct phrase **children;  /* Terminated by NULL */
};

#define svalue(va,i) (va[(i)-1].s)
#define pvalue(va,i) (va[(i)-1].p)
#define xvalue(va,i) (va[(i)-1].x)

#define att_xval(ph,f) ((ph)->va[(f)-1].x)
#define att_pval(ph,f) ((ph)->va[(f)-1].p)
#define att_sval(ph,f) ((ph)->va[(f)-1].s)

#define phrase_child(ph,att) ((ph)->va == NULL ? NULL : (ph)->va[(att)-1].p)

struct phrase *phrase_child_rec (struct phrase *p, AttId role);

#define INTERMEDIATE 1
    
#define make_phrase(CAT,I,J) \
  (TmpMATCHp = palloc(1, sizeof(struct phrase), Pool), \
   TmpMATCHp->cat = (CAT), \
   TmpMATCHp->lvl = CurrentLevel, \
   TmpMATCHp->i = (I), \
   TmpMATCHp->j = (J), \
   TmpMATCHp)


#define make_value_array()  ((union value *) palloc(NAtts, sizeof(union value), Pool) - 1)
#define make_children(n)    (palloc((n)+1, sizeof(struct phrase *), Pool))

#define SentLen (string_array_len(Words))

#define word_string(i) (string_array_ref(Words, ((i)-1)))
#define insert_word(s) (string_array_insert((s), Words))


int print_phrase (struct phrase *p, FILE *file);
int *find_elocal_positions (int cat, int start, int end);
int init_match ();

#define print_chart(file) (print_chart(1,SentLen+1,(file)))

int print_subchart (int start, int end, FILE *file);
int dump_chart (FILE *file);
int scan_sentence (FILE *file);
int match ();
int match_level (LevelId i);
struct phrase *set_children (struct phrase *p);
int *find_elocal_positions (int cat, int start, int end);
int parse_phrase (int cat, int start, int end);

#define forest_format() (print_sentence = &print_forest, PreserveSgml = 1)

int child_role (struct phrase *child, struct phrase *parent);
#define print_tree(p,file) (print_tree_1(p,NULL,file))
int print_tree_1 (struct phrase *p, struct phrase *par, FILE *file);
int print_features (struct phrase *p, FILE *file);
int print_forest ();

extern int HeadFtr;
extern int C, C0, NX, NP, NG, NMESS, PP;
/* extern int print_tuples(); */
extern int PairStyle;

extern char Tab1;
extern char Tab2;

int tuple_format ();

#define pair_format() (tuple_format(), PairStyle = 1)
#define role_format() (tuple_format(), PairStyle = 1, Tab1 = ':')
#define arg_format()  (tuple_format(), PairStyle = 1, Tab2 = ':')

extern char *TmpTUPs;
#define phrase_headword(p) \
  (p->va == NULL ? NULL : att_sval(p,HeadFtr))

/* int print_tuples (); */
/* int print_tuple (struct phrase *p, char *head, int cpos); */

#define vertical_format() (print_sentence = &print_vertical, PreserveSgml = 1)

#define level_indent(i, lvl, file) \
  for (i = NLevels; i >= lvl; i--) print_string("  ", file)

int print_tree_vertical (int r, struct phrase *p, FILE *file);
int print_preterm (int role, struct phrase *p, int width, FILE *file);
int print_vertical ();

#define binary_format() (MultiFile = 1, print_level = &write_multi)

int NodeCount;

int write_multi ();

#define multi_format() (MultiFile = 1, print_level = &print_multi)

int print_multi ();

#define phrases_format() (print_level = &print_phrases)

int print_phrases ();
int print_words (int i, int j, FILE *file);

#define xkwic_format() (MultiFile = 1, print_level = &xkwic_print_multi)

int xkwic_print_multi ();

int init_outfiles (char *filename);
int load_cass_grammar (char *grammar);

#define parse_sentence match


#endif
