#include <spa.h>
#include <random.h>

/**************************************************************************/
/**                                                                      **/
/**  distribution given as array of length n                             **/
/**                                                                      **/
/**  allow for the array to be an array of structures containing prob    **/
/**  at offset                                                           **/
/**                                                                      **/
/**  choose random 0<=i<n in accordance with prob. dist.                 **/
/**                                                                      **/
/**  let p1, ..., pn be the probabilities.  we generate a random         **/
/**  double in the range [0,1) using drand48.  The prob.dist.            **/
/**  subdivides this interval into the intervals [0,p1), ..., [pn,1).    **/
/**  We return i for the interval that the random number falls in.       **/
/**                                                                      **/
/**************************************************************************/

#define RANDMAX ((unsigned long)1<<31)
#define drandom() (((double) random())/RANDMAX)

static int ChooseInitted = 0;
int ChooseSeed;

int init_choose (int i)
{
    if (i == 0) i = time(NULL);
    ChooseSeed = i;
    srandom(i);
    ChooseInitted = 1;
}

int choose (void *dist, int n, int eltsize, int offset)
{
    double targ, d, cum = 0.0;
    char *base = (char*)dist + offset;
    int i;
    
    if (!ChooseInitted) init_choose(0);
    targ = drandom();
    for (i = 0; i < n; i++) {
	d = *(double*)(base + i*eltsize);
	cum += d;
	if (cum > targ) return i;
    }
    error("Distribution sums to less than one");
}


/**************************************************************************/
/**                                                                      **/
/**  This one builds on brandom rather than drand48, but I include it    **/
/**  here because of the conceptual connection.                          **/
/**                                                                      **/
/**  If drand48 is a better random number generator, I should rewrite    **/
/**  this sometime.                                                      **/
/**                                                                      **/
/**  This divides the interval [0,MAXINT+1) into subintervals of size    **/
/**  j-i.  If j-i does not go evenly into MAXINT+1, then the smaller     **/
/**  elements of [0,j-i) will have a slightly better chance of being     **/
/**  chosen than the larger elements.  As long as MAXINT/(j-i) is        **/
/**  large, this is negligible.                                          **/
/**                                                                      **/
/**************************************************************************/

int choose_int (int i, int j)
{
    unsigned int x; int k;

    brandom(&x, sizeof(int));
    return i + x%(j+1-i);
}
