#include <stdio.h>
#include <errno.h>
#include <spafile.h>
#include <textio.h>   /* terpri */


#include <stdlib.h>     /* for getenv() */
#include <spaconf.h>    /* for DEFAULTS_FILE */
#include <string.h>     /* for strdup() */
#include <defaults.h>

int DefaultsRead = 0;
int DefaultsVerbose = 0;

#define DONT_OVERRIDE   0
#define OVERRIDE        1
#define DRY_RUN         2
#define IGNORE_NONEXIST 4

read_default_file (char *filename, int mode)
{
    FILE *file;
    int i, len, doit, v = DefaultsVerbose || mode & DRY_RUN;
    char vbl[128], entry[1024], *old;

    file = open_file(filename, "r");
    errno = 0;
    
    if (v) {
	fprintf(stderr, "Default file: %s", filename);
	if (file == NULL) fprintf(stderr, "  -- Not found");
	terpri(stderr);
    }
    
    if (file == NULL) {
	if (!(mode & IGNORE_NONEXIST))
	  error("read_default_file: can't find file %s", filename);
    }
    else {
	while (skip_whitespace(file), scan_string(entry, 1024, "\n", file) != EOF) {
	    if (entry[0] != '#') {
		i = string_index('=', entry);
		if (i < 0 && v) {
		    fprintf(stderr, "  %s -- BAD ENTRY\n", entry);
		}
		if (i >= 0 && !(mode & DRY_RUN)) {
		    if (mode & OVERRIDE) doit = 1;
		    else {
			entry[i] = '\0';
			doit = !getenv(entry);
			entry[i] = '=';
		    }
		    if (doit) {
			putenv(strdup(entry));
			if (v) fprintf(stderr, "  %s\n", entry);
		    }
		    else if (v) fprintf(stderr, "  %s -- OVERRIDDEN\n", entry);
		}
	    }
	}
	fclose(file);
    }
}

