/**
***  Error signalling.  error() dumps core if debug_on_error is non-zero.
***  (Note: parse_args() sets debug_on_error if it receives the flag -debug.)
***  dump_core() always dumps core.
**/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>


int debug_on_error = 0;


int error (char *fmt, ...)
{
    va_list args;
  
    va_start(args,fmt);
    fprintf(stderr, "**ERROR  ");
    vfprintf(stderr, fmt, args);
    va_end(args);

    if (errno > 0) {
	fprintf(stderr, " -- %d: ", errno);
	perror(NULL);
    }
    else putc('\n', stderr);
  
    fflush(stderr);
    if (debug_on_error)
	abort();
    else
	exit(1);
}


int dump_core (char *fmt, ...)
{
    va_list args;
  
    va_start(args,fmt);
    fprintf(stderr, "**ERROR  ");
    vfprintf(stderr, fmt, args);
    va_end(args);

    if (errno > 0) {
	fprintf(stderr, " -- ");
	perror(NULL);
    }
    else putc('\n', stderr);
  
    fflush(stderr);
    errno = 0;
    abort();
}


void trace (int on, ...)
{
    va_list args;
    char *fmt;

    va_start(args,on);
    if (on) {
	fmt = va_arg(args, char*);
	vfprintf(stderr, fmt, args);
	putc('\n', stderr);
    }
    va_end(args);
}

