#include <stdio.h>
#include "spa.h"
#include "aarray.h"


/**
***  The beginning index (in lists) of the i-th list is hdr[i].
***  The length of the i-th list is hdr[i+1] - hdr[i].
**/

#define lref(a, i) \
  (aref((a)->lists, arefi((a)->hdr, (i))))

/**
***  The reason for TmpLLENj is that, technically, evaluation of
***  operands can be interleaved, so that there must be no side
***  effects in one operand that could affect the result of the
***  other.  The macro arefi does an assignment; if the two calls
***  to arefi are interleaved, we're in deep doodoo.
**/

int TmpLLENi, TmpLLENj;
#define llen(a, i) \
  (TmpLLENi = (i), \
   TmpLLENj = arefi((a)->hdr, TmpLLENi+1), \
   TmpLLENj - arefi((a)->hdr, TmpLLENi))

#define list_count(a) \
  (alen((a)->hdrs)-1)

struct list_store {
    struct aarray *hdr;
    struct aarray *lists;
};

struct list_store *make_list_store (int nlists, int nelts, int eltsize)
{
    struct list_store *s;
    int zero = 0;

    allocate(s, 1, struct list_store);
    s->hdr = make_aarray(nlists+1, sizeof(int));
    s->lists = make_aarray(nlists * nelts, eltsize);
    apush(zero, s->hdr);
    return s;
}

/**
***  The index of the new list is list_count(s) before allocation.
**/

void *lallocate (int nelts, struct list_store *s)
{
    void *list = aallocate(s->lists, nelts);

    apush(s->lists->len, s->hdr);
    return list;
}

write_list_store (struct list_store *s, FILE *file)
{
    write_aarray(s->hdr, file);
    write_aarray(s->lists, file);
}

struct list_store *read_list_store (FILE *file)
{
    struct list_store *s;

    allocate(s, 1, struct list_store);
    s->hdr = read_aarray(file);
    s->lists = read_aarray(file);
    return s;
}
