#include <stdio.h>
#include <stdarg.h>
#include "spa.h"
#include "aarray.h"
#include "textio.h"   /* terpri */

#include <global.h> /* Tmp */

typedef int ConsId;

#define make_consarray(alloc) \
  make_aarray((alloc)*2, sizeof(int))

#define carptr(c) \
  (Tmp.i = (c), Tmp.i <= 0 ? 0 : (int *) aref(Cells, Tmp.i-1))
#define car(c) *carptr(c)

#define cdrptr(c) \
  (Tmp.i = (c), Tmp.i <= 0 ? 0 : (int *) aref(Cells, Tmp.i))
#define cdr(c) *cdrptr(c)

#define cvrptr(c) \
  (Tmp.i = (c), Tmp.i <= 0 ? 0 : (int *) aref(Cells, Tmp.i+1))
#define cvr(c) *cvrptr(c)

#define caar(c) (car(car(c)))
#define cadr(c) (car(cdr(c)))
#define cdar(c) (cdr(car(c)))
#define first(c) (car(c))
#define second(c) (cadr(c))
#define push(x, loc) (Tmp.p = (loc), *(int*)Tmp.p = cons((x), *(int*)Tmp.p))

#define let_cells(cells) \
{ \
  struct aarray *let_cells_prevCells = Cells; \
  Cells = (cells);

#define end_let_cells \
  Cells = let_cells_prevCells; \
}

int cons (int a, int d);
int acons (int k, int v, int d);
int *assoc (int key, int s);
int *nassoc (int key, int *loc);
int *nassocs (int key, int *loc);
int nreverse (int s);
int list (int elt, ...);
void print_list (int a, FILE *file);
void print_alist (int a, FILE *file);
void describe_consarray (struct aarray *a, FILE *file);
void dump_aconsarray (struct aarray *a, FILE *file);



struct aarray *Cells = NULL;

ConsId cons (int a, int d)
{
    int cons;
    
    if (Cells == NULL) Cells = make_consarray(100);
    cons = apush(a, Cells) + 1;
    apush(d, Cells);
    return cons;
}

ConsId acons (int k, int v, int d)
{
    int cons;

    if (Cells == NULL) Cells = make_consarray(100);
    cons = apush(k, Cells) + 1;
    apush(d, Cells);
    apush(v, Cells);
    return cons;
}

int *assoc (int key, int s)
{
    while (s) {
	if (car(s) == key) return cvrptr(s);
	s = cdr(s);
    }
    return NULL;
}


int *nassoc (int key, int *loc)
{
    int a = *loc;

    new_entry_p = 0;
    while (a != NIL) {
	if (car(a) == key) {
	    return cvrptr(a);
	}
	a = cdr(a);
    }
    a = acons(key, NIL, *loc);
    *loc = a;
    new_entry_p = 1;
    return cvrptr(a);
}


/**
***  A little faster, but only works if the alist is sorted.
***  If you build up the alist using nassocs, it will be sorted.
**/

int *nassocs (int key, int *loc)
{
    int p = NIL, n = *loc, k, new;

    new_entry_p = 0;
    while (n != NIL) {
      k = car(n);
      if (k == key) return cvrptr(n);
      else if (k > key) break;
      p = n;
      n = cdr(n);
    }
    new = acons(key, NIL, n);
    if (p == NIL)
      *loc = new;
    else
      cdr(p) = new;
    new_entry_p = 1;
    return cvrptr(new);
}

/*
test_nassocs ()
{
  struct aarray *cells;
  int a = NIL, i;
  int entries[][2] = {
    {6, 30},
    {17, 1},
    {3, 3},
    {2, 10},
    {20, 5},
    {8, 8},
    {12, 24},
    {0, 0}
  };

  Cells = make_consarray(100);

  for (i = 0; entries[i][0]; i++) {
    *nassocs(entries[i][0], &a) = entries[i][1];
    print_alist(a, stdout);
    terpri(stdout);
  }
}
*/  
    
int *pushnew (int x, int *loc)
{
  int n = *loc;

  while (n != NIL) {
    if (car(n) == x) return loc;
    n = cdr(n);
  }
  push(x, loc);
  return loc;
}

    
int *pushnews (int x, int *loc)
{
  int p = NIL, n = *loc, k;

  while (n != NIL) {
    k = car(n);
    if (k == x) return loc;
    else if (k > x) break;
    else {
      p = n;
      n = cdr(n);
    }      
  }
  if (p == NIL)
    push(x, loc);
  else
    cdr(p) = cons(x, n);
  return loc;
}

/*
test_push ()
{
  struct aarray *cells;
  int a = NIL, i;
  int entries[] = { 6, 30, 6, 17, 17, 12, 30, 0 };

  Cells = make_consarray(100);

  for (i = 0; entries[i]; i++) {
    printf("%d -> ", entries[i]);
    pushnew(entries[i], &a);
    print_list(a, stdout);
    terpri(stdout);
  }
  terpri(stdout);

  azero(cells);
  a = NIL;
  for (i = 0; entries[i]; i++) {
    printf("%d -> ", entries[i]);
    pushnews(entries[i], &a);
    print_list(a, stdout);
    terpri(stdout);
  }

}
*/


void describe_consarray (struct aarray *a, FILE *file)
{
    fprintf(file, "struct aarray *%d used as cons array\n", a);
    fprintf(file, "  n ints = %d\n", a->len);
}

void dump_aconsarray (struct aarray *a, FILE *file)
{
    int i, *sp;
    
    if (a->len % 3 != 0) error("Not a cons array");
    fprintf(file, "      %-4s %-4s %-4s\n", "a", "v", "d");
    for (i = 0; i < a->len; i += 3) {
	sp = (int *) aref(a, i);
	fprintf(file, "%-4d: %-4d %-4d %-4d\n", i+1, sp[0], sp[2], sp[1]);
    }
}

void print_list (int a, FILE *file)
{
  int first = 1;

  fprintf(file, "(");
  while (a != NIL) {
    if (first) first = 0;
    else fprintf(file, " ");
    fprintf(file, "%d", car(a));
    a = cdr(a);
  }
  fprintf(file, ")");
}

void print_alist (int a, FILE *file)
{
  int first = 1;

  fprintf(file, "(");
  while (a != NIL) {
    if (first) first = 0;
    else fprintf(file, "\n ");
    fprintf(file, "(%d %d)", car(a), cvr(a));
    a = cdr(a);
  }
  fprintf(file, ")\n");
}

int nreverse (int s)
{
    int p = NIL, n;

    while (s != NIL) {
	n = cdr(s);
	*cdrptr(s) = p;
	p = s; s = n;
    }
    return p;
}

int list (int elt, ...)
{
    va_list elts;
    int s = NIL, n;

    va_start(elts, elt);
    if (elt != EOL) {
      s = cons(elt, NIL);
      while ((elt = va_arg(elts, int)) != EOL) {
	n = cons(elt, NIL);
	cdr(s) = n;
	s = n;
      }
    }
    return s;
}

