#include <stdio.h>
#include <global.h>
#include <types.h>
#include <allocate.h>
#include <spastrng.h>


struct object {
    TypeId type;
    union {
	char c;
	short h;
	int i;
	long l;
	float f;
	double d;
	long double D;
	char *s;
	void *p;
    } data;
};

#define type_string(t)  (TypeStrings[t])
#define type_size(t)    (TypeSizes[t])

int istype (char *s);
int type_from_size (int size);


char *TypeStringArray[] = {
    "null",
    "char",
    "short",
    "int",
    "long",
    "float",
    "double",
    "long_double",
    "string",
    "pointer",
    "fsc"
};

char **TypeStrings = &TypeStringArray[0];

int TypeSizeArray[] = {
    0,
    sizeof(char),
    sizeof(short),
    sizeof(int),
    sizeof(long),
    sizeof(float),
    sizeof(double),
    sizeof(long double),
    sizeof(char *),
    sizeof(void *),
    sizeof(void *)  /* fsc */
};

int *TypeSizes = &TypeSizeArray[0];


/**  Methods  **/

int print_short_1 ();
int print_int_1 ();
int print_long_1 ();
int print_float_1 ();
int print_double_1 ();
int print_longdbl_1 ();
int print_string_1 ();
int print_ptr_1 ();


function PrintMethods[] = {
    NULL,
    NULL,
    print_short_1,
    print_int_1,
    print_long_1,
    print_float_1,
    print_double_1,
    print_longdbl_1,
    print_string_1,
    print_ptr_1,
    NULL  /* fsc */
};


typedef void *(*read_method)(FILE *file);

void *read_fsc (FILE *file);

read_method ReadMethods[] = {
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    read_fsc
};


typedef int (*dump_method) (void *p, FILE *file);

int dump_fsc (void *p, FILE *file);

dump_method DumpMethods[] = {
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    dump_fsc
};






int istype (char *s)
{
    TypeId t;

    for (t = 0; t < NTYPES; t++) {
	if (string_equal(s, TypeStrings[t])) return t;
    }
    return 0;
}


int type_from_size (int size)
{
    int i;
    
    for (i = 0; i < NTYPES; i++) {
	if (TypeSizes[i] == size) return i;
    }
    return 0;
}


/**
***  This prints a minimal representation, definitely not multi-line, not
***  terminated by newline.
**/

int print_object (struct object *o, FILE *file)
{
    switch (o->type) {
      case T_CHAR:
	fprintf(file, "%c", o->data.c);
	break;
      case T_SHORT:
	fprintf(file, "%hd", o->data.h);
	break;
      case T_INT:
	fprintf(file, "%d", o->data.i);
	break;
      case T_LONG:
	fprintf(file, "%ld", o->data.l);
	break;
      case T_FLOAT:
	fprintf(file, "%f", o->data.f);
	break;
      case T_DOUBLE:
	fprintf(file, "%lf", o->data.d);
	break;
      case T_LONG_DOUBLE:
	fprintf(file, "%lf", o->data.D);
	break;
      default:
	if (o->type >= NTYPES) fprintf(file, "(bad type)");
	else fprintf(file, "(no print method for %s)", type_string(o->type));
	break;
    }
}


/**
***  This prints out all available information about the object in human-readable
***  form.
**/

int dump_object (struct object *o, FILE *file)
{
    int (*method)(void *, FILE *);

    method = DumpMethods[o->type];
    if (method != NULL) {
	return (*method)(o->data.p, file);
    }
    
    switch (o->type) {
      case T_CHAR:
	fprintf(file, "%c\n", o->data.c);
	break;
      case T_SHORT:
	fprintf(file, "(short) %hd\n", o->data.h);
	break;
      case T_INT:
	fprintf(file, "%d\n", o->data.i);
	break;
      case T_LONG:
	fprintf(file, "(long) %ld\n", o->data.l);
	break;
      case T_FLOAT:
	fprintf(file, "%f\n", o->data.f);
	break;
      case T_DOUBLE:
	fprintf(file, "(double) %lf\n", o->data.d);
	break;
      case T_LONG_DOUBLE:
	fprintf(file, "(long double) %lf\n", o->data.D);
	break;
      default:
	if (o->type >= NTYPES) fprintf(file, "(bad object)\n");
	else fprintf(file, "(no dump method for %s)\n", type_string(o->type));
	break;
    }
}


TypeId data_file_type (char *filename)
{
    TypeId t;
    FILE *file;

    file = fopen(filename, "r");
    if (file == NULL) error("data_file_type: can't find file %s", filename);
    read_type(t, file);
    fclose(file);
    if (t < 0 || t >= NTYPES) t = 0;
    return (TypeId) t;
}


struct object *read_object (TypeId t, FILE *file)
{
    struct object *o;
    void *(*method)(FILE *);

    allocate(o, 1, struct object);
    o->type = t;

    method = ReadMethods[t];
    if (method == NULL) error("read_object: no read method for %s", type_string(t));
    o->data.p = (*method)(file);

    return o;
}
    

struct object *load_data_file (char *filename)
{
    TypeId t;
    FILE *file;

    t = data_file_type(filename);
    file = fopen(filename, "r");
    return read_object(t, file);
}
