#ifndef REG_H
#define REG_H

#include <regfsa.h>

extern char *RegFilename;
extern char *FscFilename;



/**===================================================================
***
***				Level
***
***  This represents a level in the cascade.  The name is a category.
***  Fsr is an fsa (index), fst0 is an fsa, and nfsts tells how many
***  fsts there are for this level.  They are all sequential: if fst0
***  is 4 and nfsts is 3, then the fsts for this level are fsas 4, 5,
***  and 6.
**/

typedef int LevelId;

struct level {
    SymbolId name;   /* A cat */
    FsaId fsr;       /* FSR, merged */
    FsaId fst1;      /* First FST, one for each pattern */
    FsaId fstN;      /* Last FST (does not include the merged FSR) */
    SymbolId out1;
    SymbolId outN;
    ArrayId fstmap;  /* List: cat->fst */
};

extern struct level *Levels;
extern struct amgr *LevelMgr;

#define NLevels  (glength(LevelMgr))

#define level_name(lvl)   (Levels[lvl].name)
#define level_fsr(lvl)    (Levels[lvl].fsr)
#define level_fst1(lvl)   (Levels[lvl].fst1)
#define level_fstN(lvl)   (Levels[lvl].fstN)
#define level_out1(lvl)   (Levels[lvl].out1)
#define level_outN(lvl)   (Levels[lvl].outN)
#define level_fstmap(lvl) (Levels[lvl].fstmap)

#define new_level()       (gnew(LevelMgr))

#define init_levels()     (LevelMgr = make_amgr(&Levels, 20, sizeof(struct level)))

/* we want &fstmap[out1] = Ints + i */
#define make_fstmap(out1,nouts)  (galloc(IntMgr,(nouts)) - (out1))

extern LevelId CurrentLevel;
extern SymbolId LevelName, Out1, OutN;
extern int *FstMap;


int use_level (LevelId lvl);

extern SymbolId CatFstCat;
#define cat_fst(cat) \
  (CatFstCat = (cat), \
   CatFstCat < Out1 || CatFstCat > OutN ? 0 \
   : FstMap[CatFstCat])

LevelId find_cat_level (SymbolId cat);


/**===================================================================
***
***			      SCAN LEVEL
**/


int scan_level (struct tokstream *file);
int dump_levels (FILE *file);



/**===================================================================
***
***				 REG
***
**/


int scan_reg (struct tokstream *stream);
int write_fsc (void *p, FILE *file);
int read_fsc (FILE *file);
int free_reg ();
int load_reg (char *filename);
int load_fsc (char *filename);
int save_fsc (char *filename);
int dump_fsc (void *p, FILE *file);



/**===================================================================
***
***				 MAIN
**/


int compile_reg (char *reg_filename, char *fsc_filename);


#endif /* REG_H */
