#include <stdio.h>
#include <limits.h>
#include <spa.h>
#include <hash.h>
#include <dict.h>
#include <amgr.h>
#include <spatok.h>


#define SymbolAlloc 400
#define AttAlloc     50
#define TagAlloc   1000



/**======================================================================
***
***			       Symbols
**/

typedef int SymbolId;
typedef int AttId;

struct symbol {
    AttId att;
};

struct dict *SymbolTable = NULL;
struct symbol *Symbols = NULL;
struct amgr *SymbolMgr;

#define NullCat  1
#define NullCatString "nil"

#define find_symbol(s)       (pat_search((s), SymbolTable)+1)
#define symbol_string(cat)   (pat_key((cat)-1, SymbolTable))
#define symbol_att(sym)      (Symbols[sym].att)

#define NSymbols (glength(SymbolMgr))

int init_regdict ();
SymbolId intern (char *s);



/**===================================================================
***
***			      Tags
**/

typedef int TagId;
typedef char TType;

#define INPUT        0
#define ACTION       1
#define OUTPUT       2
#define TAGBASE      3

#define ACT_LEFT     5
#define ACT_RIGHT    9
#define ACT_ASSIGN  13


struct tag {
    SymbolId symbol;
    TType type;
};

struct tag *Tags = NULL;
struct amgr *TagMgr;
struct hash_table *TagTable = NULL;

char *TagTypeString[] = { "I", "E", "O", 0, 0, "L", 0, 0, 0, "R", 0, 0, 0, "A" };
char *TagTypePrettyString[] =    {  "", "-", "$", 0, 0, "[", 0, 0, 0, "]", 0, 0, 0, "@" };

#define tag_symbol(t)             (Tags[t].symbol)
#define tag_type(t)               (Tags[t].type)
#define tag_string(t)             (symbol_string(Tags[t].cat))
#define tt_string(tt)             (TagTypeString[tt])
#define tt_pstring(tt)            (TagTypePrettyString[tt])

#define new_tag()                 ((TagId) gnew(TagMgr))

#define NTags (glength(TagMgr))

#define init_tags() \
  (TagMgr = make_amgr(&Tags, TagAlloc, sizeof(struct tag)), \
   TagTable = make_hash_table(TagAlloc, sizeof(int)+1, sizeof(int)))


int clear_tags ();
TagId intern_tag (SymbolId sym, TType type);



/**===================================================================
***
***			       Attributes
**/

typedef char AType;
typedef char AMode;
typedef int RangeId;

struct att {
    SymbolId symbol;
    AMode mode;
    AType type;
    int field;
    RangeId range;
};

struct att *Atts = NULL;
struct amgr *AttMgr;

#define NAtts (glength(AttMgr))

#define intern_att(s)      (need_att(intern(s)))
#define att_symbol(a)      (Atts[a].symbol)
#define att_string(a)      (symbol_string(att_symbol(a)))
#define att_type(a)        (Atts[a].type)
#define att_mode(a)        (Atts[a].mode)
#define att_field(a)       (Atts[a].field)
#define att_range(a)       (Atts[a].range)

#define amode_string(m)    (AttModeStrings[m])
#define atype_string(t)    (AttTypeStrings[t])

#define new_att()          ((AttId) gnew(AttMgr))

#define ATT_STRING 1
#define ATT_PHRASE 2
#define ATT_SUBSET 3

#define ATT_SET    0
#define ATT_AND    1
#define ATT_IOR    2
#define ATT_XOR    3

char *AttTypeStrings[] = { "NULL", "string", "phrase", "subset" };
char *AttModeStrings[] = { "=",  "&",  "|",  "^" };


AttId need_att (SymbolId sym);
int scan_att_decls (struct tokstream *file);


/**====================================================================
***
***			      Range Sets
**/

/**  A range set is a sequence of symbols.  We represent a range set  **/
/**  as an offset in the array RangeSets.  The offset points to the   **/
/**  0th element of the range set, which is its length, n.  The 1 to  **/
/**  nth elements are the SymbolId's of the symbols of the range set. **/

typedef int RangelId;
typedef unsigned long t_subset;

struct range {
    int n;
    SymbolId cat1;
    SymbolId catN;
    RangelId cats;
    RangelId map;
};

struct range *Ranges = NULL;
struct amgr *RangeMgr;

int *Rangels = NULL;
struct amgr *RangelMgr;

#define NRanges        (glength(RangeMgr))
#define NextRangel     (glength(RangelMgr) + 1)
#define NRangels       (glength(RangelMgr))

#define range_n(r)     (Ranges[r].n)
#define range_cat1(r)  (Ranges[r].cat1)
#define range_catN(r)  (Ranges[r].catN)
#define range_cats(r)  (Ranges[r].cats)
#define range_map(r)   (Ranges[r].map)

#define range_sym(r,i)     ((Rangels+range_cats(r))[i])
#define range_index(r,sym) ((SymbolId) (Rangels+range_map(r))[sym])

RangeId scan_range (struct tokstream *file);



/**====================================================================
***
***			      Functions
**/


int init_regdict ()
{
    SymbolTable = make_dict(SymbolAlloc);
    SymbolMgr = make_amgr(&Symbols, SymbolAlloc, sizeof(struct symbol));
    if (intern(NullCatString) != NullCat) {
	error("init_regdict: this shouldn't happen (NullCat)");
    }
    AttMgr = make_amgr(&Atts, AttAlloc, sizeof(struct att));
}


int clear_tags ()
{
    if (Tags == NULL) init_tags();
    else {
	gzero(TagMgr);
	clrhash(TagTable);
    }
}


SymbolId intern (char *s)
{
    SymbolId i;

    if (SymbolTable == NULL) init_regdict();
    i = (SymbolId) pat_insert((s), SymbolTable) + 1;
    if (new_entry_p) {
	if (gnew(SymbolMgr) != i) error("intern: SymbolTable and Symbols out of synch");
    }
    return i;
}


TagId intern_tag (SymbolId sym, TType type)
{
    TagId t;

    if (Tags == NULL) init_tags();

    if ((type & TAGBASE) == ACTION) {
	t = new_tag();
	Tags[t].symbol = sym;
	Tags[t].type = type;
	if (type == ACT_ASSIGN) need_att(sym);
	return t;
    }
    else {
	char key[sizeof(int)+1];
	int *tp;

	bcopy(&sym, key, sizeof(int));
	bcopy(&type, key+sizeof(int), 1);
	tp = needhash(key, TagTable);
	if (new_entry_p) {
	    t = new_tag();
	    *tp = t;
	    Tags[t].symbol = sym;
	    Tags[t].type = type;
	}
	return *tp;
    }
}


AttId need_att (SymbolId s)
{
    AttId a;

    if (Symbols == NULL) init_regdict();

    a = Symbols[s].att;
    if (a == 0) {
	Symbols[s].att = a = new_att();
	Atts[a].symbol = s;
    }
    else {
	if (Atts[a].symbol != s) error("need_att: symbol att disagrees with att symbol");
    }
    return a;
}


int scan_att_decls (struct tokstream *file)
{
    char *token = file->token;
    int tt, t;
    SymbolId sym;
    AttId att;

    while (1) {

	tt = scan_token(file);
	if (tt == EOF) break;
	if (tt == COLON) {
	    unscan_token(tt, file);
	    break;
	}

	/* att type */
	if (tt != STRING) {
	    error("%s, line %d, token %s: missing attribute type",
		  file->filename, file->line, toktype_string(tt));
	}
	if (string_equal(token, "string")) t = ATT_STRING;
	else if (string_equal(token, "phrase")) t = ATT_PHRASE;
	else if (string_equal(token, "subset")) t = ATT_SUBSET;
	else error("%s, line %d: Unrecognized attribute type %s",
		   file->filename, file->line, token);

	/* attribute name */
	tt = scan_token(file);
	if (tt == EOF) error("%s: premature EOF", file->filename);
	if (tt != STRING) {
	    error("%s, line %d, token %s: expecting attribute name",
		  file->filename, file->line, toktype_string(tt));
	}
	att = intern_att(token);
	if (!new_entry_p) {
	    error("%s, line %d: redefining attribute %s",
		  file->filename, file->line, token);
	}
	att_type(att) = t;
	att_mode(att) = t==ATT_SUBSET ? ATT_AND : ATT_SET;
	att_field(att) = t==ATT_STRING ? 1 : t==ATT_SUBSET ? 3 : 0;

	/* field & range */
	while ((tt = scan_token(file)) == COMMA) {
	    tt = scan_token(file);
	    if (tt != STRING) {
		error("%s, line %d, token %s: expecting attribute modifier",
		      file->filename, file->line, toktype_string(tt));
	    }
	    if (string_equal(token, "field")) {
		tt = scan_token(file);
		if (tt != STRING) {
		    error("%s, line %d, token %s: expecting numeric value for field",
			  file->filename, file->line, toktype_string(tt));
		}
		att_field(att) = atoi(token);
	    }
	    else if (string_equal(token, "range")) {
		att_range(att) = scan_range(file);
	    }
	    else {
		error("%s, line %d: unrecognized attribute modifier %s",
		      file->filename, file->line, token);
	    }
	}
	if (tt != SEMICOLON) {
	    error("%s, line %d, token %s: expecting semicolon at end of attribute",
		  file->filename, file->line, toktype_string(tt));
	}
    }

    return 1;
}


RangeId scan_range (struct tokstream *file)
{
    int tt, elt, elt1, n, *map;
    RangeId r;
    char *token = file->token;
    SymbolId cat, cat1, catN, *cats;
    
    if (Ranges == NULL) {
	RangeMgr = make_amgr(&Ranges, 20, sizeof(struct range));
	RangelMgr = make_amgr(&Rangels, 100, sizeof(int));
    }

    tt = scan_token(file);
    if (tt != LBRACE) {
	error("%s, line %d, token %s: expecting left brace for range set",
	      file->filename, file->line, toktype_string(tt));
    }

    elt1 = NextRangel;
    cat1 = INT_MAX;
    catN = INT_MIN;
    while ((tt = scan_token(file)) == STRING) {
	elt = gnew(RangelMgr);
	if (elt1 == 0) elt1 = elt;
	cat = intern(token);
	Rangels[elt] = cat;
	if (cat < cat1) cat1 = cat;
	if (cat > catN) catN = cat;
    }
    n = NextRangel - elt1;
    r = gnew(RangeMgr);
    range_n(r) = n;
    range_cat1(r) = cat1;
    range_catN(r) = catN;
    range_cats(r) = elt1 - 1;
    range_map(r) = galloc(RangelMgr, n) - cat1;
    cats = Rangels + range_cats(r);
    map = Rangels + range_map(r);

    for (elt = 1; elt <= n; elt++) {
	cat = cats[elt];
	map[cat] = elt;
    }

    if (tt != RBRACE) {
	error("%s, line %d, token %s: expecting right brace after range set",
	      file->filename, file->line, toktype_string(tt));
    }

    return r;
}


t_subset atosubset (char *s, RangeId r)
{
    t_subset x;
    int i, n, *map;
    char *t;
    SymbolId cat;
    
    x = 0;
    n = range_n(r);
    map = Rangels + range_map(r);
    while (*s) {
	for (t = s; *t && *t != '|'; t++) ;
	if (*t == '|') {
	    *t = 0;
	    t++;
	}
	cat = find_symbol(s);
	if (cat < range_cat1(r) || cat > range_catN(r))
	  error("atosubset: symbol %s not in range", s);
	i = map[cat];
	if (i == 0) error("atosubset: symbol %s not in range", s);
	x |= 1 << i-1;
	s = t;
    }
    return x;
}


int print_subset (t_subset s, RangeId r, FILE *file)
{
    int i, n, first;
    t_subset x;
    SymbolId *cats;

    n = range_n(r);
    cats = Rangels + range_cats(r);
    x = 1;
    first = 1;
    for (i = 1; i <= n; i++, x <<= 1) {
	if (x & s) {
	    if (!first) print_char('|', file);
	    else first = 0;
	    print_string(symbol_string(cats[i]), file);
	}
    }
}


int sprint_subset (t_subset s, RangeId r, char *str, int len)
{
    char *tmp;
    int i, n, first;
    t_subset x;
    SymbolId *cats;

    len--;  /* for the terminating null */
    n = range_n(r);
    cats = Rangels + range_cats(r);
    x = 1;
    first = 1;
    for (i = 1; i <= n; i++, x <<= 1) {
	if (x & s) {
	    if (!first) {
		if (len <= 0) {*str = 0; return 1;}
		*str++ = '|';
		len--;
	    }
	    else first = 0;
	    tmp = symbol_string(cats[i]);
	    while (*tmp) {
		if (len <= 0) {*str = 0; return 1;}
		*str++ = *tmp++;
		len--;
	    }
	}
    }
    *str = 0;
    return 1;
}


int print_range (RangeId r, FILE *file)
{
    int *cats, i, n;

    n = range_n(r);
    cats = Rangels + range_cats(r);

    fprintf(file, "{");
    for (i = 1; i <= n; i++) {
	fprintf(file, " %s", symbol_string(cats[i]));
    }
    fprintf(file, " }");
}


int dump_regdict (FILE *file)
{
    SymbolId sym;
    TagId tag;
    AttId a;
    RangeId r;
    int i;

    fprintf(file, "Symbols:\nsym att string\n");
    for (sym = 1; sym <= NSymbols; sym++) {
	fprintf(file, "%3d %3d %s\n", sym, symbol_att(sym), symbol_string(sym));
    }
    terpri(file);

    if (Tags != NULL) {
	fprintf(file, "Tags:\ntag type sym\n");
	for (tag = 1; tag <= NTags; tag++) {
	    fprintf(file, "%3d %3d %3d %s%s\n",
		    tag, tag_type(tag), tag_symbol(tag),
		    tt_pstring(tag_type(tag)), symbol_string(tag_symbol(tag)));
	}
	terpri(file);
    }

    if (Atts != NULL) {
	fprintf(file, "Attributes:\natt   type field mode range string\n");
	for (a = 1; a <= NAtts; a++) {
	    fprintf(file, "%3d %6s %5d %4s %5d %s\n",
		    a, atype_string(att_type(a)), att_field(a),
		    amode_string(att_mode(a)), att_range(a), att_string(a));
	}
	terpri(file);
    }

    if (Ranges != NULL) {
	fprintf(file, "Ranges:\n");
	fprintf(file, "    r   n cat1 catN cats map\n");
	for (r = 1; r <= NRanges; r++) {
	    fprintf(file, "  %3d %3d %4d %4d %4d %3d ",
		    r, range_n(r), range_cat1(r), range_catN(r), range_cats(r), range_map(r));
	    print_range(r, file);
	    terpri(file);
	}
	terpri(file);
	fprintf(file, "Rangels:\n");
	fprintf(file, "         1   2   3   4   5   6   7   8   9  10");
	for (i = 1; i <= NRangels; i++) {
	    if (i%10 == 1) fprintf(file, "\n  %3d:", i-1);
	    fprintf(file, " %3d", Rangels[i]);
	}
	terpri(file);
    }
}


#define NoSymbols    1
#define NoAttributes 2
#define NoRanges     4

write_regdict (FILE *file)
{
    char flags = 0;

    if (Symbols == NULL) SET(flags, NoSymbols);
    if (Atts == NULL) SET(flags, NoAttributes);
    if (Ranges == NULL) SET(flags, NoRanges);
    write_char(flags, file);

    if (Symbols != NULL) {
	write_dict(SymbolTable, file);
	write_amgr(SymbolMgr, file);
    }

    if (Atts != NULL) {
	write_amgr(AttMgr, file);
    }

    if (Ranges != NULL) {
	write_amgr(RangeMgr, file);
	write_amgr(RangelMgr, file);
    }
}

read_regdict (FILE *file)
{
    char flags;

    flags = read_char(file);

    if (!GET(flags, NoSymbols)) {
	SymbolTable = read_dict(file);
	SymbolMgr = read_amgr(&Symbols, file);
    }

    if (!GET(flags, NoAttributes)) {
	AttMgr = read_amgr(&Atts, file);
    }

    if (!GET(flags, NoRanges)) {
	RangeMgr = read_amgr(&Ranges, file);
	RangelMgr = read_amgr(&Rangels, file);
    }
}

free_regdict ()
{
    if (Symbols != NULL) {
	free_dict(SymbolTable);
	SymbolTable = NULL;
	free_amgr(SymbolMgr);
	SymbolMgr = NULL;
	Symbols = NULL;
    }
    if (Tags != NULL) {
	free_amgr(TagMgr);
	TagMgr = NULL;
	Tags = NULL;
	free_hash_table(TagTable);
	TagTable = NULL;
    }
    if (Atts != NULL) {
	free_amgr(AttMgr);
	AttMgr = NULL;
	Atts = NULL;
    }
    if (Ranges != NULL) {
	free_amgr(RangeMgr);
	RangeMgr = NULL;
	Ranges = NULL;
	free_amgr(RangelMgr);
	RangelMgr = NULL;
	Rangels = NULL;
    }
}

