#ifndef REGDICT_H
#define REGDICT_H


#include <spa.h>
#include <hash.h>
#include <dict.h>
#include <aarray.h>
#include <spatok.h>


typedef int SymbolId;
typedef int AttId;

struct symbol {
    AttId att;
};

extern struct dict *SymbolTable;
extern struct symbol *Symbols;
extern struct amgr *SymbolMgr;

#define NullCat  1
#define NullCatString "nil"

#define find_symbol(s)       (pat_search((s), SymbolTable)+1)
#define symbol_string(cat)   (pat_key((cat)-1, SymbolTable))
#define symbol_att(sym)      (Symbols[sym].att)


int init_regdict ();
SymbolId intern (char *s);


typedef int TagId;
typedef signed char TType;

#define INPUT        0
#define ACTION       1
#define OUTPUT       2
#define TAGBASE      3

#define ACT_LEFT     5
#define ACT_RIGHT    9
#define ACT_ASSIGN  13


struct tag {
    SymbolId symbol;
    TType type;
};

extern struct tag *Tags;
extern struct amgr *TagMgr;
extern struct hash_table *TagTable;

#define NTags (glength(TagMgr))

extern char *TagTypeString[];
extern char *TagTypePrettyString[];

#define tag_symbol(t)             (Tags[t].symbol)
#define tag_type(t)               (Tags[t].type)
#define tag_string(t)             (symbol_string(Tags[t].symbol))
#define tt_string(tt)             (TagTypeString[tt])
#define tt_pstring(tt)            (TagTypePrettyString[tt])

#define new_tag()                 ((TagId) gnew(TagMgr))

int clear_tags ();
TagId intern_tag (SymbolId sym, TType type);


typedef char AType;
typedef char AMode;
typedef int RangeId;

struct att {
    SymbolId symbol;
    AMode mode;
    AType type;
    int field;
    RangeId range;
};

extern struct att *Atts;
extern struct amgr *AttMgr;

#define NAtts (glength(AttMgr))

#define intern_att(s)      (need_att(intern(s)))
#define att_symbol(a)      (Atts[a].symbol)
#define att_string(a)      (symbol_string(att_symbol(a)))
#define att_type(a)        (Atts[a].type)
#define att_mode(a)        (Atts[a].mode)
#define att_field(a)       (Atts[a].field)
#define att_range(a)       (Atts[a].range)

#define amode_string(m)    (AttModeStrings[m])
#define atype_string(t)    (AttTypeStrings[t])

#define new_att()          ((AttId) gnew(AttMgr))

#define ATT_STRING 1
#define ATT_PHRASE 2
#define ATT_SUBSET 3

#define ATT_SET    0
#define ATT_AND    1
#define ATT_IOR    2
#define ATT_XOR    3

extern char *AttTypeStrings[];
extern char *AttModeStrings[];


AttId need_att (SymbolId sym);
int scan_att_decls (struct tokstream *file);





typedef int RangelId;
typedef unsigned long t_subset;

struct range {
    int n;
    SymbolId cat1;
    SymbolId catN;
    RangelId cats;
    RangelId map;
};

extern struct range *Ranges;
extern struct amgr *RangeMgr;

extern int *Rangels;
extern struct amgr *RangelMgr;

#define NRanges        (glength(RangeMgr))
#define NextRangel     (glength(RangelMgr) + 1)
#define NRangels       (glength(RangelMgr))

#define range_n(r)     (Ranges[r].n)
#define range_cat1(r)  (Ranges[r].cat1)
#define range_catN(r)  (Ranges[r].catN)
#define range_cats(r)  (Ranges[r].cats)
#define range_map(r)   (Ranges[r].map)

#define range_sym(r,i)     ((Rangels+range_cats(r))[i])
#define range_index(r,sym) ((SymbolId) (Rangels+range_map(r))[sym])

RangeId scan_range (struct tokstream *file);

int print_range (RangeId r, FILE *file);
t_subset atosubset (char *s, RangeId r);
int print_subset (t_subset s, RangeId r, FILE *file);
int sprint_subset (t_subset s, RangeId r, char *str, int len);

#define subset_memq(sym,S,r) ((S) & 1<<range_index((r),(sym))-1)

int dump_regdict (FILE *file);

#endif
