#ifndef REGFSA_H
#define REGFSA_H


#include <regdict.h>
#include <regparse.h>



/**-------------------------------------------------------------------
***
***			      Reg Arrays
**/

extern int *Ints;
extern struct amgr *IntMgr;

typedef int ArrayId;


extern char *Chars;
extern struct amgr *CharMgr;


/**-------------------------------------------------------------------
***
***				 FSA
***
***  This structure contains all the permanent information about FSA's.
***  Information needed only during compilation is kept in a regex
***  structure.
***
***  The lists are indices of subarrays in the aarray Lists.
**/

typedef int FsaId;

struct fsa {
    SymbolId cat;
    int nlocals;     /* the number of locals */
    int ilocals;     /* the number of input locals (the rest are epsilon locals) */
    int nstates;     /* the number of states */
    SymbolId mincat;
    SymbolId maxcat;
    ArrayId cmap;    /* maps cat - cat0 to a local or -1 */
    ArrayId lmap;    /* maps local to cat */
    ArrayId emap;    /* maps local to type */
    ArrayId tm;      /* transition matrix */
    ArrayId finals;  /* array of cats */
};

extern struct fsa *Fsas;
extern struct amgr *FsaMgr;

#define NFsas (glength(FsaMgr))

extern FsaId CurrentFsa;
extern SymbolId MinCat, MaxCat;
extern int NLocals, ILocals;
extern int *CMap, *LMap, *TM, *Finals, *TMap;
extern char *EMap;
extern struct amgr *TMapMgr;


int use_fsa (FsaId fsa);
int clear_fsa ();


#define new_fsa()    (gnew(FsaMgr))
#define NextFsa      (glength(FsaMgr)+1)

#define fsa_cat(i)       (Fsas[i].cat)
#define fsa_name(i)      (symbol_string(fsa_cat(i)))
#define fsa_nlocals(i)   (Fsas[i].nlocals)
#define fsa_ilocals(i)   (Fsas[i].ilocals)
#define fsa_nstates(i)   (Fsas[i].nstates)
#define fsa_mincat(i)    (Fsas[i].mincat)
#define fsa_maxcat(i)    (Fsas[i].maxcat)
#define fsa_cmap(i)      (Fsas[i].cmap)
#define fsa_lmap(i)      (Fsas[i].lmap)
#define fsa_emap(i)      (Fsas[i].emap)
#define fsa_tm(i)        (Fsas[i].tm)
#define fsa_finals(i)    (Fsas[i].finals)

int print_fsa (FsaId i, FILE *file);
int print_fsa_maps (struct fsa *fsa, FILE *file);
int print_fsa_transition_matrix (struct fsa *fsa, FILE *file);

extern SymbolId CatLocalTmp;
#define cat_local(cat) \
  (CatLocalTmp = (cat), \
   CatLocalTmp < MinCat || CatLocalTmp > MaxCat ? 0 \
   : CMap[CatLocalTmp])

#define local_cat(lcl)        (LMap[lcl])
#define elocal_type(lcl)      (EMap[lcl])
#define local_type(lcl)       (lcl <= ILocals ? INPUT : elocal_type(lcl))
#define tag_local(tag)        (TMap[tag])

char *local_string (int local);



/**--------------------------------------------------------------------
***
***				States
**/

extern int RegTraceTM;

extern struct hash_table *States;
extern int StateHashAlloc;
extern int Nbits, Nlits, Lit0, NStates;

#define start_state         1
#define state_table(s)      (TM + (s) * NLocals)
#define state_next1(s,lcl)  (state_table(s)[lcl])

extern SymbolId StateNextTmp;
#define state_next(s,cat) \
  (StateNextTmp = (cat), \
   StateNextTmp < MinCat || StateNextTmp > MaxCat ? 0 : \
   state_next1(s, CMap[StateNextTmp]))

extern int StateFinalsTmp;
#define state_finals(s) \
  (StateFinalsTmp = Finals[s], \
   StateFinalsTmp ? Ints + StateFinalsTmp : NULL)



/**===================================================================
***
***			      Functions
**/


int use_fsa (FsaId fsa);
int clear_fsa ();
FsaId compile_regex (RegexId rx);
int init_fsas ();
int write_fsas (FILE *file);
int read_fsas (FILE *file);
int free_fsas ();
int print_fsa (FsaId fsa, FILE *file);
int dump_fsas (FILE *file);
int dump_fsa_stores (FILE *file);
int dump_fsa_ints (FILE *file);
int dump_fsa_chars (FILE *file);
int make_maps (FsaId fsa, RegexId rx);
int print_maps (FILE *file);
int init_states (RegexId rx);
int intern_state (int lits, struct tstore *tstore);
int make_transition_matrix (FsaId fsa, RegexId rx);
int print_tm (FILE *file);


#endif
