#ifndef REGPARSE_H
#define REGPARSE_H

#include <amgr.h>
#include <hash.h>
#include <rbtree.h>
#include <dict.h>
#include <regdict.h>
#include <spatok.h>



/* extending type list from spatok.h */

#define TERM  NTOKTYPES
#define SEQ   NTOKTYPES+1
#define ALT   NTOKTYPES+2
#define EXP   NTOKTYPES+3

#define NREGTYPES  NTOKTYPES+4

extern int RegTraceRX;
extern int PrintCompileMessages;

#define STACKSIZE 128
#define ESTACKSIZE 32


typedef int RegexId;
typedef int PointId;
typedef int LiteralId;
typedef int ListId;


/**===================================================================
***
***				Regex
**/

struct regex {
    SymbolId cat;
    int flags;
    PointId lp;
    PointId rp;
    PointId pt1;
    PointId ptN;
    LiteralId lit1;
    LiteralId litN;
};

extern struct regex *Regices;
extern struct amgr *RegexMgr;

#define new_regex()    (gnew(RegexMgr))

#define regex_cat(i)   (Regices[i].cat)
#define regex_flags(i) (Regices[i].flags)
#define regex_lp(i)    (Regices[i].lp)
#define regex_rp(i)    (Regices[i].rp)
#define regex_pt1(i)   (Regices[i].pt1)
#define regex_ptN(i)   (Regices[i].ptN)
#define regex_lit1(i)  (Regices[i].lit1)
#define regex_litN(i)  (Regices[i].litN)
#define regex_fsa(i)   (Regices[i].fsa)

#define NextRegex      (glength(RegexMgr) + 1)
#define NRegices       (glength(RegexMgr))

#define RX_MACRO       1
#define RX_NO_ACTIONS  2
#define RX_DELETED     4

#define regex_is_macro(rx)     (GET(regex_flags(rx), RX_MACRO))
#define regex_with_actions(rx) (!GET(regex_flags(rx), RX_NO_ACTIONS))
#define regex_is_deleted(rx)   (GET(regex_flags(rx), RX_DELETED))



/**-------------------------------------------------------------------
***
***				Points
**/

struct point {
    ListId lits;
    PointId to1;
    PointId to2;
};

extern struct point *Points;
extern struct amgr *PointMgr;

extern struct tstore *PointTstore;
extern struct tstore *TodoTstore;
extern struct tstore *AuxTstore;

#define new_point()    (gnew(PointMgr))
#define NextPoint      (glength(PointMgr)+1)
#define NPoints        (glength(PointMgr))

#define point_lits(p)  (Points[p].lits)
#define point_to1(p)   (Points[p].to1)
#define point_to2(p)   (Points[p].to2)

#define print_points(file) (print_point_range(1, NPoints, (file)))

#define print_regex_points(file) \
  (print_point_range(Regex->pt1, Regex->ptN, (file)))



/**-------------------------------------------------------------------
***
***			       Literals
**/

struct literal {
    TagId tag;
    PointId rp;
};

extern struct literal *Literals;
extern struct amgr *LiteralMgr;

#define new_literal()   (gnew(LiteralMgr))
#define NextLiteral     (glength(LiteralMgr)+1)
#define NLiterals       (glength(LiteralMgr))

#define lit_tag(lit)    (Literals[lit].tag)
#define lit_rp(lit)     (Literals[lit].rp)
#define lit_string(lit) (tag_string(lit_tag(lit)))

#define print_literals(file) (print_literal_range(1,NLiterals,(file)))


/**-------------------------------------------------------------------
***
***				Nodes
**/

struct node {
    int t;
    SymbolId v;
};


/**===================================================================
***
***			      Functions
**/


char *type_name (int t);
int init_regices ();
RegexId find_regex (SymbolId cat);

#define RegTraceSyntax 1
#define RegTraceDump   2

RegexId scan_regex (struct tokstream *file);
int expand_macro (SymbolId cat, PointId *lpp, PointId *rpp);
int propagate (ListId todo);
int delete_actions (RegexId rx);
RegexId combine_regices (SymbolId cat, RegexId rx1, RegexId rxN);
int print_regex (RegexId r, FILE *file);
int dump_regices (FILE *file);
int print_point_range (PointId start, PointId end, FILE *file);
int print_literal (int lit, FILE *file);
int print_literal_range (int start, int end, FILE *file);


#endif
