#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <defaults.h>
#include <spafile.h>
#include <spastrng.h>   /* string_equal */


FILE *TmpSPAfile = NULL;
int FopenErrNo = 0;


FILE *Fopen(char *path, char *mode)
{
  FILE *f;

  if (path == NULL || strcmp(path, "-") == 0)
    return (mode == NULL || mode[0] == 'r') ? stdin : stdout;

  if ((f = fopen(path, mode)) == NULL) {
    fprintf(stderr, "Error opening %s (fopen)\n", path);
    perror("");
    exit(1);
  }
  return f;
}


int file_exists_p (char *filename)
{
    struct stat s;
    int ok = stat(filename, &s);

    if (ok == 0) return 1;
    else if (errno == ENOENT ||
	errno == ENOTDIR) {
	errno = 0;
	return 0;
    }
    else return 0;
}


int file_size (char *filename)
{
    struct stat s;

    if (stat(filename, &s) < 0)
      error("file_size: Can't stat %s", filename);
    return (int) s.st_size;
}


time_t file_mtime (char *filename)
{
    struct stat s;

    if (stat(filename, &s) < 0)
      error("file_mtime: Can't stat %s", filename);
    return s.st_mtime;
}


int rm (char *filename)
{
    if (unlink(filename) < 0) {
	/* OK if file doesn't exist */
	if (errno == ENOENT)
	  errno = 0;
	else error("rm: Unable to delete %s", filename);
    }
}



int catenate (char **filenames, int nfiles, FILE *outfile)
{
    char buf[BLOCKSIZE];
    int n;
    FILE *file;

    for (; nfiles > 0; filenames++, nfiles--) {
	file = fopen(*filenames, "r");

	if (file == NULL)
	  fprintf(stderr, "catenate: can't open %s, skipping\n", *filenames);
	else {
	    while ((n = fread(buf, sizeof(char), BLOCKSIZE, file)) == BLOCKSIZE)
	      fwrite(buf, sizeof(char), BLOCKSIZE, outfile);
	    if (n > 0)
	      fwrite(buf, sizeof(char), n, outfile);
	}
    }
}



char *dir_next (DIR *dir)
{
  struct dirent *e;

  e = readdir(dir);
  if (e == NULL) return NULL;
  return e->d_name;
}


int isabsolute (char *filename)
{
    return (filename[0] == '/' ||
	    filename[0] == '.' && filename[1] == '/' ||
	    filename[0] == '.' && filename[1] == '.' && filename[2] == '/');
}


char *DataPath1 = NULL;


/**  filename should point to at least MAXFILENAME bytes of memory  **/

char *find_filename (char *filename, char *path)
{
    char tmp[MAXFILENAME], pathbuf[1024], *p = pathbuf;
    int dlen, flen;
    struct stat statbuf;

    flen = strlen(filename);

    if (isabsolute(filename)) {
	if (DefaultsVerbose) fprintf(stderr, "Absolute filename\n");
	if (stat(filename, &statbuf) >= 0) return filename;
	else return NULL;
    }

    if (DefaultsVerbose) fprintf(stderr, "Searching Path %s\n", path);

    /**  two nulls at the end, the extra one for stopping the search  **/
    if ((dlen = strlen(path))+2 > 1024) error("find_file: path buffer overflow");
    strcpy(p, path);
    p[dlen+1] = '\0';

    for (; *p; p += dlen+1) {
	for (dlen = 0; p[dlen] != ':' && p[dlen] != '\0'; dlen++);
	p[dlen] = '\0';
	if (dlen+flen+2 > MAXFILENAME) error("find_filename: pathname buffer overflow");
	sprintf(tmp, "%s/%s", p, filename);
	if (DefaultsVerbose) fprintf(stderr, "Trying %s\n", tmp);
	if (stat(tmp, &statbuf) >= 0) {
	    strcpy(filename, tmp);
	    return filename;
	}
	errno = 0;
    }
    return NULL;
}


char *filesuf (char *filename)
{
    int i = strlen(filename);

    if (i > 0) {
	while (--i) {
	    if (filename[i] == '.') return filename + i + 1;
	}
    }
    return NULL;
}


int has_suf (char *filename, char *suf)
{
    int len, slen;

    len = strlen(filename);
    slen = strlen(suf);
    if (slen >= len) return 0;
    filename += len-slen-1;
    if (filename[0] == '.' && string_equal(filename+1, suf))
      return 1;
    else return 0;
}


char *addsuf (char *filename, char *suf)
{
    int len = strlen(filename), suflen = strlen(suf);

    if (!(len > suflen && string_equal(filename + len - suflen, suf))) {
	if (len+suflen+1 > MAXFILENAME) error("addsuf: filename overflow");
	strcpy(filename + len, suf);
    }
    return filename;
}


char *chsuf (char *filename, char *oldsuf, char *newsuf)
{
    int len = strlen(filename), suflen = strlen(oldsuf);

    if (len > suflen && string_equal(filename + len - suflen, oldsuf))
      len -= suflen;
    suflen = strlen(newsuf);
    if (!(len > suflen && string_equal(filename + len - suflen, newsuf))) {
	if (len+suflen+1 > MAXFILENAME) error("chsuf: filename overflow");
	strcpy(filename + len, newsuf);
    }
    return filename;
}


char *rmsuf (char *filename, char *suf)
{
    int len = strlen(filename), suflen = strlen(suf);

    if (len > suflen && string_equal(filename + len - suflen, suf)) {
	filename[len-suflen] = '\0';
    }
    return filename;
}
