#include <stdio.h>
#include <spastrng.h>
#include <allocate.h>
#include <defaults.h>


char *StringSpace = NULL;
int StringSpaceFree = 0;


char *copy_string (char *s)
{
    int len = strlen(s) + 1;

    if (len > StringSpaceFree) {
	if (len <= 32) {
	    allocate(StringSpace, 1024, char);
	    StringSpaceFree = 1024;
	}
	else {
	    char *new;
	    allocate(new, len, char);
	    strcpy(new, s);
	    return new;
	}
    }
    strcpy(StringSpace, s);
    s = StringSpace;
    StringSpace += len;
    StringSpaceFree -= len;
    return s;
}


int string_index (char c, char *s)
{
    int i;

    for (i = 0; s[i] != '\0'; i++)
      if (s[i] == c) return i;
    return -1;
}


int string_fields (char *s, char **a, int len, char *delims, int squeeze)
{
    int k = 0, i = 0, j = 0;

    while (1) {
	j = i;
	while (s[j] != '\0' && !string_member(s[j], delims)) j++;
	if (squeeze == 0 || j > i) {
	    if (k >= len) error("Overflowed array in string_fields");
	    a[k++] = s + i;
	}
	if (s[j] == '\0') return k;
	s[j] = '\0';
	i = j + 1;
    }
}


int is_number (char *s)
{
    int type = ISINT;
    
    for (; *s; s++) {
	if (!(isdigit(*s) || *s == '-' || *s == '+' || *s == '.'))
	  return 0;
	if (*s == '.') type = ISFLOAT;
    }
    return type;
}


void downcase_string (char *s)
{
    while (*s) {
	*s = downcase_char(*s);
	s++;
    }
}


void upcase_string (char *s)
{
    while (*s) {
	*s = upcase_char(*s);
	s++;
    }
}


char *expand_string (char *s)
{
    static char buf[1024];
    char *v, spc;
    int i, j, o = 0, len;

    for (i = 0; s[i]; i++) {
	if (o >= 1024) {
	    fprintf(stderr, "expand_string: buffer overflow\n");
	    exit(1);
	}

	/** Dollar Sign **/
	if (s[i] == '$') {
	    i++;

	    /** Dollar Escape **/
	    if (s[i] == '$') {
		buf[o++] = '$';
	    }

	    /** Variable Substitution **/
	    else {
		if (s[i] == '{' || s[i] == '(') {
		    i++;
		    for (j = i; s[j] && s[j] != '}' && s[j] != ')'; j++);
		    spc=0;
		}
		else {
		    for (j = i; s[j] && (s[j] == '_' || isalnum(s[j])); j++);
		    spc = s[j];
		}
		s[j] = '\0';
		v = getdefault(s+i);
		if (v != NULL) {
		    len = strlen(v);
		    if (o+len >= 1024) {
			fprintf(stderr, "expand_string: buffer overflow\n");
			exit(1);
		    }
		    strcpy(buf+o, v);
		    o += len;
		}
		i = j;
		if (spc) s[i--] = spc;
	    }
	}

	/** Backslash Escape **/
	else if (s[i] == '\\') {
	    i++;
	    if (s[i] == 'n') {
		buf[o++] = '\n';
	    }
	    else if (s[i] == 't') {
		buf[o++] = '\t';
	    }
	    else buf[o++] = s[i];
	}
	    
	else buf[o++] = s[i];
    }
    return buf;
}
