#include <stdio.h>
#include <spa.h>
#include <strtab.h>


#define MAX_STEM 64

#define NONE	0
#define ADD_E	1
#define I_TO_Y	2
#define DEGEM	3

#define SUFF_S	1
#define SUFF_D	2
#define SUFF_G	3

char *Decode_suffix[8] = { NULL, "S", "D", "G", "N", "d", "R", "M" };

#define decode_suffix(suff) (Decode_suffix[suff])

int Encode_suffix[52] = {

    0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/*  a  b  c  d  e  f  g  h  i  j  k  l  m  n  o  p  q  r  s  t  u  v  w  x  y  z  */

    0, 0, 0, 2, 0, 0, 3, 0, 0, 0, 0, 0, 7, 4, 0, 0, 0, 6, 1, 0, 0, 0, 0, 0, 0, 0
/*  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  P  Q  R  S  T  U  V  W  X  Y  Z  */

    };

int encode_suffix(char suff)
{
  if (suff >= 'a' && suff <= 'z')
    return Encode_suffix[suff-'a'];
  if (suff >= 'A' && suff <= 'Z')
    return Encode_suffix[suff-'A'+26];
  return 0;
}

static char stem[MAX_STEM+1];
static int stemlen, suffix, change;


int strip_suffix_return (int len, int suff, int chng)
{
  stemlen -= len;
  if (stemlen < 1) {
    fprintf(stderr, "Stripped too much (%s)\n", stem);
    exit(1);
  }
  stem[stemlen] = '\0';
  suffix = suff;
  change = chng;
  return chng;
}


#define vowel_p_in_context(s, i, last)         \
  ((!consonant_p(s[i]) &&                          \
    /* u is not a vowel following q */             \
    (s[i] != 'u' || i == 0 || s[i-1] != 'q')) ||   \
   /* y is a vowel between two consonants */       \
   (s[i] == 'y' &&                                 \
    (i <= 0 || consonant_p(s[i-1])) &&             \
    (i >= last || consonant_p(s[i+1]))))


int monosyllable_p (char *s, int i, int j)
{
  int vowel = 0, last = strlen(s)-1;

  for (; i < j; i++) {
    /* if there are two vowels, it is not a monosyllable */
    if (vowel_p_in_context(s, i, last)) {
      if (vowel)
	return 0;
      vowel = 1;
    }
  }
  return 1;
}


int consonant_cluster_p (char *s, int i, int j)
{
  int last = strlen(s)-1;
  
  for (; i < j; i++) {
    if (vowel_p_in_context(s, i, last)) {
      return 0;
    }
  }
  return 1;
}


/* assumption: suffix==0
 * side effect: stripping suffix, setting suffix
 * value: whether or not to fix the stem
 */

int strip_suffix ()
{
  int c1 = stem[stemlen-1];
  int c2 = stemlen<2 ? 0 : stem[stemlen-2];
  int c3 = stemlen<3 ? 0 : stem[stemlen-3];
  int c4 = stemlen<4 ? 0 : stem[stemlen-4];

  switch (c1) {

  case 's':

    if (!c2) return 0;

    switch (c2) {

      /* ss */
    case 's': return 0;

    case 'e':

      if (consonant_cluster_p(stem, 0, stemlen-2)) return 0;

      /* [oiS]es */
      switch (c3) {
      case 'o': case 'i': case 's': case 'z': case 'x': case 'h':
	return strip_suffix_return(2, SUFF_S, 1);

	/* es */
      default: return strip_suffix_return(1, SUFF_S, 0);
      }

    case 'u':

      /* eaus */
      if (c4 == 'e' && c3=='a')
	return strip_suffix_return(1, SUFF_S, 0);

      /* us */
      return 0;

      /* is */
    case 'i': return 0;

      /* s */
    default: return strip_suffix_return(1, SUFF_S, 0);
    }

  case 'd':
    if (c2!='e') return 0;

    /* Ked */
    if (consonant_cluster_p(stem, 0, stemlen-2)) return 0;

    /* eed */
    if (c3=='e') return 0;

    /* Cred */
    if (consonant_p(c4) && c4 != 'r' && c3=='r')
      return 0;

    /* ed */
    return strip_suffix_return(2, SUFF_D, 1);

  case 'g':
    if(c2!='n' || c3!='i') return 0;

    /* King */
    if (!(stemlen > 4 && c4 == 'y') &&              /* crying, drying, etc. */
	consonant_cluster_p(stem, 0, stemlen-3))
      return 0;

    /* Cring */
    if (stemlen>=5 &&
	c4 == 'r' &&
	stem[stemlen-5] != 'r' &&
	consonant_p(stem[stemlen-5]))
      return 0;

    /* ing */
    return strip_suffix_return(3, SUFF_G, 1);

  case 'n':

    /* men */
    if (c3 == 'm' &&
	c2 == 'e') {
      stem[stemlen-2] = 'a';
      suffix = SUFF_S;
      return 0;
    }
  }

  return 0;
}


/*  Assumes that c has been downcased */

int consonant_p (int c)
{
  static int consonants[] = {
    0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1
    };

  if (c==0) return 0;

  c = downcase_char(c);
  return c >= 'a' && c <=  'z' ? consonants[c - 'a'] : 0;
}


int stem_change ()
{
  int c1 = stem[stemlen-1];
  int c2 = stemlen<2 ? 0 : stem[stemlen-2];
  int c3 = stemlen<3 ? 0 : stem[stemlen-3];
  int c4 = stemlen<4 ? 0 : stem[stemlen-4];

  switch (c1) {

    /* Vowels */
  case 'a': return NONE;
  case 'o': return NONE;
  case 'i': return I_TO_Y;
  case 'e': return NONE;
  case 'u': return ADD_E;

    /* xx; x */
  case 'x':
    if (c2=='x') return DEGEM;
    else return NONE;

    /* tz, zz; z */
  case 'z':
    if (c2=='t' || c2=='z') return NONE;
    else return ADD_E;

    /* ss; s */
  case 's':
    if (c2=='s') return NONE;
    else return ADD_E;

    /* et, it */
  case 't':
    if (c2=='e' || c2=='i') return NONE;
    else break;

    /* vv, gg, cc; v, g, c */
  case 'v':
  case 'g':
  case 'c':
    if (c1==c2) return DEGEM;
    else return ADD_E;

    /* ff */
  case 'f':
    if (c2=='f') return NONE;
    else break;

    /* wl, rl; Mll; ual, ial */
  case 'l':
    if (c2=='w' ||
	c2=='r' ||
	(c2 == 'l' && monosyllable_p(stem, 0, stemlen)) ||
	c2=='e' ||
	(c2=='a' && (c3=='u' || c3=='i')))
      return NONE;

    /* ll */
    else if (c2=='l') return DEGEM;

    /* Cl */
    else if (consonant_p(c2)) return ADD_E;
    else break;

    /* rr; Cr */
  case 'r':
    if (c2=='r') return DEGEM;
    else if (consonant_p(c2)) return ADD_E;
    else break;

    /* th */
  case 'h':
    if (c2=='t') return ADD_E;
    else break;

    /* ^.y, ^.w; y, w */
  case 'y':
  case 'w':
    if (stemlen==2) return ADD_E;
    else return NONE;

    /* VCick, VCCick */
  case 'k':
    if (stemlen>=5 &&
	c2=='c' &&
	c3=='i' &&
	consonant_p(c4) &&
	(!consonant_p(stem[stemlen-5]) ||
	 stemlen>5 && !consonant_p(stem[stemlen-6])))
      return DEGEM;
    else break;
  }

  /** C **/

  if (consonant_p(c2)) {

    /* XX */
    if (c1==c2) return DEGEM;

    /* CyC */
    /* (we know c1 is a consonant because we exhausted the vowels above) */
    if (c2=='y' && consonant_p(c3)) return ADD_E;

    /* CC */
    return NONE;
  }

  /** VC **/

  /* iaC, uaC, uiC, uoC */
  if ((c3=='i' && c2=='a') ||
      (c3=='u' && (c2=='a' || c2=='i' || c2=='o')))
    return ADD_E;

  /* VVC */
  if (!consonant_p(c2) && !consonant_p(c3)) return NONE;

  /** CVC **/

  /* em, om, en, on, er, or */
  if ((c2=='e' || c2=='o') &&
      (c1=='m' || c1=='n' || c1=='r'))
    return NONE;

  /* CVC */
  return ADD_E;
}


do_change (int change)
{
  switch (change) {

  case ADD_E:
    stem[stemlen]='e';
    stem[stemlen+1]='\0';
    break;

  case I_TO_Y:
    stem[stemlen-1]='y';
    break;

  case DEGEM:
    stem[stemlen-1]='\0';
    break;
  }
}


struct string_table *word_table = NULL;
struct string_table *stem_table = NULL;
    
    
void load_word_exceptions (char *path)
{
  FILE *file = Fopen(path, "r");
  word_table = scan_string_table(1000, file);
  fclose(file);
}

void load_stem_exceptions (char *path)
{
  FILE *file = Fopen(path, "r");
  stem_table = scan_string_table(1000, file);
  fclose(file);
}


char *look_up_word (char *word)
{
  char *v, *s;

  if (word_table == NULL) return 0;
  if (!(v = string_table_ref(word, word_table))) return 0;
  s = stem;
  while (*v && *v != ' ')
    *s++ = *v++;
  *s = '\0';
  if (!*v)
    suffix = NONE;
  else if (!*(v+1) || *(v+2)) {
    fprintf(stderr, "Bad suffix in word exception table: \"%s\" \"%s\"\n", stem, v);
    exit(1);
  }
  else {
    suffix = encode_suffix(*++v);
    if (!suffix) {
      fprintf(stderr, "Unknown suffix in word exception table: \"%s\" \"%s\"\n",
	      stem, v);
      exit(1);
    }
  }
  return stem;
}

char *look_up_stem (char *s)
{
  char *v;


  if (!stem_table) return 0;
  if (!(v = string_table_ref(s, stem_table))) return 0;
  strcpy(stem, v);
  return stem;
}


char *stem_word (char *word)
{
  char tmp[MAX_STEM];

  strcpy(tmp, word);
  if (look_up_word(tmp)) return stem;
  downcase_string(tmp);
  if (look_up_word(tmp)) {
    /* This doesn't get the case with funny capitalization */
    stem[0] = upcase_char(stem[0]);
    return stem;
  }

  strcpy(stem, word);
  stemlen = strlen(stem);
  suffix = change = 0;
  strip_suffix();

  if (change) {
    strcpy(tmp, stem);
    if (look_up_stem(tmp)) return stem;
    downcase_string(tmp);
    if (look_up_stem(tmp)) {
      stem[0] = upcase_char(stem[0]);
      return stem;
    }
    do_change(stem_change());
  }

  return stem;
}


int type;

void stem_text (FILE *file)
{
  char word[MAX_STEM], *w;
  int c;

  w = word;
  *word = '\0';
  while ((c = scan_char(file)) != EOF) {
    if (isalpha(c))
      *w++ = c;
    else {
      if (*word) {
	*w = '\0';
	stem_word(word);
	fputs(stem, stdout);
	w = word;
	*word = '\0';
      }
      putchar(c);
    }
  }
}


void stem_words (FILE *file)
{
  char word[MAX_STEM];

  while (fscanf(file, "%s", word) == 1) {
    stem_word(word);
    if (suffix)
      printf("%s %s\n", stem, decode_suffix(suffix));
    else printf("%s\n", stem);
  }
}   


int stemmer_main (int argc, char *argv[])
{
  int i, text = 0;
  char *inputfn = NULL, filename[MAXFILENAME], *we = NULL, *se = NULL;
  FILE *file;

  parse_args(&argc, &argv, "?s-sw-ss-bx", &inputfn, &we, &se, &text);

  if (we != NULL) load_word_exceptions(we);
  if (se != NULL) load_stem_exceptions(se);

  if (word_table == NULL) {
      strcpy(filename, "stemmer_words");
      if (find_datafile(filename) == NULL)
	error("%s: %s not found", ProgName, filename);
      file = Fopen(filename, "r");
      word_table = scan_string_table(1000, file);
      fclose(file);
  }
  if (stem_table == NULL) {
      strcpy(filename, "stemmer_stems");
      if (find_datafile(filename) == NULL)
	error("%s: %s not found", ProgName, filename);
      file = Fopen(filename, "r");
      stem_table = scan_string_table(1000, file);
      fclose(file);
  }

  file = (inputfn == NULL ? stdin : Fopen(inputfn, "r"));
  if (text)
    stem_text(file);
  else stem_words(file);

  return 0;
}
