#include <stdio.h>
#include <spa.h>
#include <dict.h>
#include <hash.h>



#define WORDS_CASE_SENSITIVE 1
#define DOWNCASE_INPUT_TAGS  2
#define INSERT_EOS           4


struct tagfixes {
    int flags;
    struct dict *dict;         /* words & cats */
    struct hash_table *hash;   /* cat word-code -> new-cat */
};


int TagFixesAlloc = 1000;
int KeepOldTags = 0;
int WordsCaseSensitive = 0;
int DowncaseInputTags = 0;
int TagfixesSilent = 0;
int InsertEos = 0;
char *EosSym;
int IsEos;
struct tagfixes *TagFixes = NULL;


struct tagfixes *make_tagfixes (int alloc)
{
    struct tagfixes *fx;

    allocate(fx, 1, struct tagfixes);
    if (WordsCaseSensitive) SET(fx->flags, WORDS_CASE_SENSITIVE);
    if (DowncaseInputTags) SET(fx->flags, DOWNCASE_INPUT_TAGS);
    if (InsertEos) SET(fx->flags, INSERT_EOS);
    fx->dict = make_dict(alloc);
    fx->hash = make_hash_table(alloc, sizeof(int)*2, sizeof(int));
    return fx;
}


#define keyw  (key[0])
#define keyt  (key[1])

#define ntfld (r.fields[0])
#define wfld  (r.fields[1])
#define tfld  (r.fields[2])

struct tagfixes *scan_tagfixes (FILE *file)
{
    int c, newtag, key[2];
    struct tagfixes *fx;
    struct record r;

    /* Sets global variables */
    scan_tagfix_options(file);

    fx = make_tagfixes(TagFixesAlloc);

    if (pat_insert("", fx->dict) != 0)
      error("scan_tagfixes: first dict index is supposed to be 0");

    if (InsertEos) {
	if (pat_insert(EosSym, fx->dict) != 1)
	  error("scan_tagfixes: eos dict index is supposed to be 1");
    }

    while (scan_record_1(&r, file, " \t", SQUEEZE) != EOF) {
	if (r.nfields > 0 && r.fields[0][0] != '#') {
	    if (r.nfields != 3)
	      fprintf(stderr, "scan_tagfixes: Warning: nfields neq 3, skipping\n");
	    else {
		newtag = pat_insert(ntfld, fx->dict);

		if (!WordsCaseSensitive) downcase_string(wfld);
		if (string_equal(wfld, "*")) keyw = 0;
		else keyw = pat_insert(wfld, fx->dict);

		if (string_equal(tfld, "*")) keyt = 0;
		else keyt = pat_insert(tfld, fx->dict);

		sethash(&key, &newtag, fx->hash);
	    }
	}
    }
    return fx;
}


int scan_tagfix_options (FILE *file)
{
    char vbl[128], val[128];
    int delim;
    
    skip_whitespace(file);
    if (peek_char(file) != '{') return 0;
    scan_char(file);
    while (skip_whitespace(file), peek_char(file) != '}') {
	delim = scan_string(vbl, 128, " \t\n=", file);
	if (delim != '=') {
	    skip_whitespace(file);
	    if (scan_char(file) != '=')
	      error("scan_tagfix_options: missing equals after option");
	}
	downcase_string(vbl);
	skip_whitespace(file);
	scan_string(val, 128, " \t\n", file);
	if (string_equal(vbl, "words_case_sensitive"))
	  WordsCaseSensitive = atoi(val);
	else if (string_equal(vbl, "downcase_input_tags"))
	  DowncaseInputTags = atoi(val);
	else if (string_equal(vbl, "eos")) {
	    InsertEos = 1;
	    EosSym = strdup(val);
	}
	else error("Unrecognized option: %s", vbl);
    }
    scan_char(file);   /* closing brace */
}
	  

int compile_tagfixes (char *origfilename)
{
    char filename[MAXFILENAME];
    FILE *infile, *outfile;
    struct tagfixes *fx;

    strcpy(filename, origfilename);
    if (has_suf(filename, "fxc"))
      error("compile_tagfixes: input file can't have suffix .fxc (%s)", filename);
    infile = fopen(filename, "r");
    if (infile == NULL) {
	if (has_suf(filename, "fx"))
	  error("compile_tagfixes: can't find file %s", filename);
	else {
	    addsuf(filename, ".fx");
	    infile = fopen(filename, "r");
	    if (infile == NULL)
	      error("compile_tagfixes: can't find file %s[.fx]", origfilename);
	}
    }
    chsuf(filename, ".fx", ".fxc");
    outfile = Fopen(filename, "w");

    fx = scan_tagfixes(infile);
    write_tagfixes(fx, outfile);

    fclose(infile);
    fclose(outfile);

    if (!TagfixesSilent) printf("Wrote %s\n", filename);
}


char *fix_tag (char *word, char *tag)
{
    int t, w, *tp, key[2];

    IsEos = 0;

    t = pat_search(tag, TagFixes->dict);
    w = pat_search(word, TagFixes->dict);

    if (InsertEos && t == 1) IsEos = 1;

    if (w == 0) {
	fprintf(stderr, "fix_tag: Warning: empty word\n");
	return tag;
    }
    if (t == 0) {
	fprintf(stderr, "fix_tag: Warning: empty word\n");
	return tag;
    }

    if (w > 0 && t > 0) {
	keyt = t;
	keyw = w;
	tp = gethash(key, TagFixes->hash);
	if (tp != NULL) return pat_key(*tp, TagFixes->dict);
    }
    if (w > 0) {
	keyt = 0;
	keyw = w;
	tp = gethash(key, TagFixes->hash);
	if (tp != NULL) return pat_key(*tp, TagFixes->dict);
    }
    if (t > 0) {
	keyt = t;
	keyw = 0;
	tp = gethash(key, TagFixes->hash);
	if (tp != NULL) return pat_key(*tp, TagFixes->dict);
    }
    return tag;
}


#undef wfld
#define wfld  (r.fields[0])
#undef tfld
#define tfld  (r.fields[1])

int apply_tagfixes (struct tagfixes *fx, FILE *infile, FILE *outfile)
{
    int k;
    char *tag;
    struct record r;
    struct tagfixes *prev = TagFixes;

    TagFixes = fx;
    WordsCaseSensitive = GET(fx->flags, WORDS_CASE_SENSITIVE);
    DowncaseInputTags = GET(fx->flags, DOWNCASE_INPUT_TAGS);
    InsertEos = GET(fx->flags, INSERT_EOS);
    IsEos = InsertEos;

    while (scan_record(&r, infile) != EOF) {
	if (IsEos) fprintf(outfile, "<s>\n");
	if (r.nfields >= 2) {
	    print_string(wfld, outfile);
	    if (KeepOldTags) fprintf(outfile, "\t%s", tfld);
	    if (!WordsCaseSensitive) downcase_string(wfld);
	    if (DowncaseInputTags) downcase_string(tfld);
	    fprintf(outfile, "\t%s", fix_tag(wfld, tfld));
	    for (k = 2; k < r.nfields; k++) {
		fprintf(outfile, "\t%s", r.fields[k]);
	    }
	    fprintf(outfile, "\n");
	    if (IsEos) fprintf(outfile, "</s>\n");
	}
	else if (r.nfields == 1)
	  fprintf(outfile, "%s\n", r.fields[0]);
    }
    TagFixes = prev;
}


write_tagfixes (struct tagfixes *fx, FILE *file)
{
    write_int(fx->flags, file);
    write_dict(fx->dict, file);
    write_hash_table(fx->hash, file);
}


struct tagfixes *read_tagfixes (FILE *file)
{
    struct tagfixes *fx;

    allocate(fx, 1, struct tagfixes);
    read_int(fx->flags, file);
    fx->dict = read_dict(file);
    fx->hash = read_hash_table(file);
    return fx;
}


struct tagfixes *load_tagfixes (char *name)
{
    char filename[MAXFILENAME];
    FILE *file;
    struct tagfixes *fx;

    strcpy(filename, name);
    chsuf(filename, ".fx", ".fxc");
    if (find_datafile(filename) != NULL) {
	file = Fopen(filename, "r");
	fx = read_tagfixes(file);
	fclose(file);
	return fx;
    }

    chsuf(filename, ".fxc", ".fx");
    if (find_datafile(filename) != NULL) {
	file = Fopen(filename, "r");
	fx = scan_tagfixes(file);
	fclose(file);
	return fx;
    }
	
    error("load_tagfixes: can't find %s", name);
}
