#include <stdio.h>
#include <allocate.h>
#include <spastrng.h>
#include <spafile.h>


#include <errno.h>

#define read_char getc
#define write_char putc
#define scan_char getc
#define print_char fputc

#define unread_char ungetc
#define unscan_char ungetc
int peek_char (FILE *stream);

#define terpri(file) fprintf(file, "\n")

int read_string (char *s, int n, FILE *file);
char *read_string_alloc (FILE *file);
int write_string (char *s, FILE *file);
int scan_string (char *s, int n, char *delims, FILE *file);
#define print_string fputs
int print_string_1 (char **s, FILE *file);

int scan_line (char *s, int alloc, FILE *file);

int skip_chars (char *chars, FILE *stream);
int skip_chars_not (char *chars, FILE *stream);
#define skip_whitespace(file) (skip_chars(" \t\n\r\f\v", (file)))

int skip_whitespace_cl (FILE *file);
#define skip_line(file) (skip_chars_not("\n", (file)), scan_char(file))


struct lined_file {
    char *filename;
    FILE *file;
    long int line;
};

struct lined_file *lopen (char *filename);
int lread_char (struct lined_file *file);
int lunread_char (char c, struct lined_file *file);
#define lpeek_char(file) (peek_char((file)->file))


#define RECORD_LINE_ALLOC 1024
#define RECORD_FIELDS_ALLOC 128

struct record {
    char line[RECORD_LINE_ALLOC];
    char *fields[RECORD_FIELDS_ALLOC];
    int nfields;
};

#define scan_record(r, file) scan_record_1(r, file, "\t", DONT_SQUEEZE)
int scan_record_1 (struct record *rec, FILE *file, char *delims, int squeeze);
int print_record (struct record *r, FILE *file);

int print_field (int n, char *s, FILE *file);
int indent (FILE *file);
int iterpri (FILE *file);

#define MAX_INDENTS 64

#define push_indent(i) \
  (++IndentsTop >= MAX_INDENTS ? error("push_indent: stack overflow") : 0, \
   Indents[IndentsTop] = (IndentsTop == 0 ? 0 : Indents[IndentsTop-1]), \
   Indents[IndentsTop] += (i))

#define pop_indent() \
  (IndentsTop < 0 ? error("pop_indent: empty stack") : IndentsTop--)

#define clear_indent() (IndentsTop = -1)


/**-------------------------------------------------------------------
***  Functions
***/  


int peek_char (FILE *stream)
{
  int c;
  c = read_char(stream);
  if (c != EOF) unread_char(c, stream);
  return c;
}


int read_string (char *s, int n, FILE *file)
{
    int c, i = 0;

    while ((c = read_char(file)) != EOF && c != '\0') {
	if (i >= n) dump_core("Buffer overflow reading string");
	s[i++] = c;
    }
    if (i == 0 && c == EOF) return EOF;
    if (i >= n) dump_core("Buffer overflow reading string");
    s[i] = '\0';
    return i;
}


char *read_string_alloc (FILE *file)
{
    char *s, buf[1024];
    int len;

    if ((len = read_string(buf, 1024, file)) == EOF)
      return NULL;
    else {
	allocate(s, len + 1, char);
	strcpy(s, buf);
	return s;
    }
}


int write_string (char *s, FILE *file)
{
    int i = 0;
    while (s[i] != '\0')
      write_char(s[i++], file);
    write_char('\0', file);
    return i;
}


int scan_string (char *s, int n, char *delims, FILE *file)
{
    int c, i = 0;

    while ((c = read_char(file)) != EOF && !string_member(c, delims)) {
	if (i >= n) dump_core("Buffer overflow scanning string");
	s[i++] = c;
    }
    if (i == 0 && c == EOF) return EOF;
    if (i >= n) dump_core("Buffer overflow scanning string");
    s[i] = '\0';
    if (c != EOF) unscan_char(c, file);
    return (c == EOF ? 0 : c);
}


int scan_line (char *s, int n, FILE *file)
{
    int c, i = 0;

    while ((c = read_char(file)) != EOF && c != '\n') {
	if (i >= n) error("Buffer overflow scanning string");
	s[i++] = c;
    }
    if (i == 0 && c == EOF) return EOF;
    if (i >= n) error("Buffer overflow scanning string");
    s[i] = '\0';
    return 1;
}


int print_string_1 (char **s, FILE *file)
{
    print_string(*s, file);
}


int skip_chars (char *chars, FILE *stream)
{
  int c;
  while ((c = fgetc(stream)) != EOF && string_member(c, chars) != NULL);
  if (c != EOF) ungetc(c, stream);
  return 1;
}


int skip_chars_not (char *chars, FILE *stream)
{
  int c;
  while ((c = fgetc(stream)) != EOF && string_member(c, chars) == NULL);
  if (c != EOF) ungetc(c, stream);
  return 1;
}

int LineNumber = 1;

   
skip_whitespace_cl (FILE *file)
{
    while (1) {
	skip_chars(" \t", file);
	if (peek_char(file) != '\n') break;
	scan_char(file);
	LineNumber++;
    }
}


struct lined_file *lopen (char *filename)
{
    struct lined_file *file;

    allocate(file, 1, struct lined_file);

    if (filename == NULL) {
	file->filename = NULL;
	file->file = stdin;
    }
    else {
	file->filename = strdup(filename);
	file->file = Fopen(filename, "r");
    }
    file->line = 1;
    return file;
}

int lclose (struct lined_file *file)
{
    if (file->filename != NULL) {
	free(file->filename);
	fclose(file->file);
    }
    free(file);
}

int lread_char (struct lined_file *file)
{
    int c;

    c = read_char(file->file);
    if (c == '\n') file->line++;
    return c;
}

int lunread_char (char c, struct lined_file *file)
{
    if (c == '\n') file->line--;
    unread_char(c, file->file);
}

int scan_record_1 (struct record *rec, FILE *file, char *delims, int squeeze)
{
    if (scan_line(rec->line, RECORD_LINE_ALLOC, file) == EOF)
      return EOF;
    rec->nfields = string_fields(rec->line, rec->fields, RECORD_FIELDS_ALLOC, delims, squeeze);
    return 1;
}


int print_record (struct record *r, FILE *file)
{
  int i;
  
  for (i = 0; i < r->nfields; i++) {
    if (i != 0) print_char(' ', file);
    print_string(r->fields[i], file);
  }
}


print_field (int n, char *s, FILE *file)
{
    char buffer[128];
    int i, lws = 0, rws = 0, sgn = n < 0 ? -1 : 1;

    n *= sgn;
    if (n >= 128) error("print_field: buffer overflow");
    if (sgn < 0) lws = n - strlen(s);
    else rws = n - strlen(s);
    strncpy(buffer+lws, s, n);
    buffer[n] = '\0';
    for (i = 0; i < lws; i++)
      buffer[i] = ' ';
    for (i = n-rws; i < n; i++)
      buffer[i] = ' ';
    print_string(buffer, file);
}


int Indents[MAX_INDENTS];
int IndentsTop = -1;


indent (FILE *file)
{
    int i;

    if (IndentsTop >= 0) {
	for (i = 0; i < Indents[IndentsTop]; i++)
	  print_char(' ', file);
    }
}


iterpri (FILE *file)
{
    int i;

    print_char('\n', file);
    for (i = 0; i < Indents[IndentsTop]; i++)
      print_char(' ', file);
}
