#ifndef TEXTIO_H
#define TEXTIO_H

#include <errno.h>

#define read_char getc
#define write_char putc
#define scan_char getc
#define print_char fputc

#define unread_char ungetc
#define unscan_char ungetc
int peek_char (FILE *stream);

#define terpri(file) fprintf(file, "\n")

int read_string (char *s, int n, FILE *file);
char *read_string_alloc (FILE *file);
int write_string (char *s, FILE *file);
int scan_string (char *s, int n, char *delims, FILE *file);
#define print_string fputs
int print_string_1 (char **s, FILE *file);

int scan_line (char *s, int alloc, FILE *file);

int skip_chars (char *chars, FILE *stream);
int skip_chars_not (char *chars, FILE *stream);
#define skip_whitespace(file) (skip_chars(" \t\n\r\f\v", (file)))
extern int LineNumber;
int skip_whitespace_cl (FILE *file);
#define skip_line(file) (skip_chars_not("\n", (file)), scan_char(file))


struct lined_file {
    char *filename;
    FILE *file;
    long int line;
};

struct lined_file *lopen (char *filename);
int lread_char (struct lined_file *file);
int lunread_char (char c, struct lined_file *file);
#define lpeek_char(file) (peek_char((file)->file))
int lclose (struct lined_file *file);

#define RECORD_LINE_ALLOC 1024
#define RECORD_FIELDS_ALLOC 128

struct record {
    char line[RECORD_LINE_ALLOC];
    char *fields[RECORD_FIELDS_ALLOC];
    int nfields;
};

#define scan_record(r, file) scan_record_1(r, file, "\t", DONT_SQUEEZE)
int scan_record_1 (struct record *rec, FILE *file, char *delims, int squeeze);
int print_record (struct record *r, FILE *file);

int print_field (int n, char *s, FILE *file);
int indent (FILE *file);
int iterpri (FILE *file);

#define MAX_INDENTS 64

extern int Indents[MAX_INDENTS];
extern int IndentsTop;

#define push_indent(i) \
  (++IndentsTop >= MAX_INDENTS ? error("push_indent: stack overflow") : 0, \
   Indents[IndentsTop] = (IndentsTop == 0 ? 0 : Indents[IndentsTop-1]), \
   Indents[IndentsTop] += (i))

#define pop_indent() \
  (IndentsTop < 0 ? error("pop_indent: empty stack") : IndentsTop--)

#define clear_indent() (IndentsTop = -1)

#endif
