#include <spa.h>
#include <aarray.h>
#include <dict.h>


struct xdict {
    struct dict *d;
    struct aarray *a;
};


int xdict_insert (char *key, struct xdict *D, void **entp)
{
    int i;
    void *p;

    i = pat_insert(key, D->d);
    p = aneed(D->a, i);
    if (entp != NULL) *entp = p;
    return i+1;
}


int xdict_search (char *key, struct xdict *D, void **entp)
{
    int i;

    i = pat_search(key, D->d);
    if (i < 0) return 0;
    else {
	if (entp != NULL) *entp = aref(D->a, i);
	return i+1;
    }
}


#define xdict_key(id, D) (pat_key((id)-1, (D)->d))
#define xdict_ref(D, id) (aref((D)->a, (id)-1))


struct xdict *make_xdict (int alloc, int eltsize)
{
    struct xdict *D;

    allocate(D, 1, struct xdict);
    D->d = make_dict(alloc);
    D->a = make_aarray(alloc, eltsize);
    return D;
}


int reset_xdict (struct xdict *D)
{
    free_dict(D->d);
    D->d = make_dict(D->a->alloc);
    azero(D->a);
}


int free_xdict (struct xdict *D)
{
    free_dict(D->d);
    free_aarray(D->a);
    free(D);
    return 1;
}


int write_xdict (struct xdict *D, FILE *file)
{
    write_dict(D->d, file);
    write_aarray(D->a, file);
    return 1;
}


struct xdict *read_xdict (FILE *file)
{
    struct xdict *D;

    allocate(D, 1, struct xdict);
    D->d = read_dict(file);
    D->a = read_aarray(file);
    return D;
}
