#include <spa.h>
#include <cass.h>


main (int argc, char *argv[])
{
    char *infilename = NULL, *outfilename = NULL, *grammar = NULL, *fxfile = NULL;
    char *format = NULL, filename[MAXFILENAME], *grepcat = NULL;
    int fx = 0, vert = 0, phrases = 0, nokids = 0, where = 0;
    int noatts = 0, yesatts = 0, noroles = 0, yesroles = 0, nocps = 0, yescps = 0;
    FILE *infile = stdin;

    defaults(format, "CASS_FORMAT");
    defaults(grammar, "CASS_GRAMMAR");
    defaults(Head, "CASS_HEADATT");
    defaulti(DontPrintAttributes, "CASS_DONT_PRINT_ATTRIBUTES");
    defaulti(DontPrintCPs, "CASS_DONT_PRINT_CPS");
    defaulti(DontPrintRoles, "CASS_DONT_PRINT_ROLES");
    defaults(Bos, "CASS_BOS");
    defaults(Eos, "CASS_EOS");

    Usage = "cass [options] [infile]\n\
Infile defaults to stdin\n\
-g default is value of `CASS_GRAMMAR'\n\
-o defaults to stdout\n\
-X means no actions: no feature passing or inserted phrases\n\
-s means silent (no warnings)\n\
\n\
-format\n\
   legal values are: forest [default], vertical, multi, binary, xkwic, phrases\n\
   if not provided, takes value of env var `CASS_FORMAT'\n\
-v is shorthand for `-format vertical'\n\
-p is shorthand for `-format phrases'\n\
-grep <cat>\n\
\n\
-A or CASS_DONT_PRINT_ATTRIBUTES suppresses printing of attributes\n\
-a causes printing of attributes\n\
-N or CASS_DONT_PRINT_CPS suppresses printing of corpus positions\n\
-n causes printing of corpus positions\n\
-R or CASS_DONT_PRINT_ROLES suppresses printing of roles\n\
-r causes printing of roles\n\
-ch suppresses children in phrases format\n\
-bos <string> to set the BOS marker\n\
-eos <string> to set the EOS marker\n\
\n\
-debug causes a trace of all parser actions to be dumped to stdout.";

    parse_args(&argc, &argv,
	       "?s-sg-sf-so-bdebug-bX-bs-sformat-bv-bp-bch-sh-bw-bA-ba-bN-bn-bR-br-sgrep-sbos-seos",
	       &infilename, &grammar, &grammar, &outfilename, &MatchVerbose, &NoActions,
	       &NoWarn, &format, &vert, &phrases, &nokids, &Head, &where,
	       &noatts, &yesatts, &nocps, &yescps, &noroles, &yesroles, &grepcat,
	       &Bos, &Eos);

#ifndef DEBUG
    if (MatchVerbose) printf("Sorry, not compiled DEBUG, no verbose");
#endif
    if (phrases) format = "phrases";
    if (vert) format = "vertical";
    if (nokids) PrintChildren = 0;
    if (noatts) DontPrintAttributes = 1;
    if (yesatts) DontPrintAttributes = 0;
    if (nocps) DontPrintCPs = 1;
    if (yescps) DontPrintCPs = 0;
    if (noroles) DontPrintRoles = 1;
    if (yesroles) DontPrintRoles = 0;

    /**  Grammar  **/
    if (grammar == NULL) {
	grammar = &filename[0];
	strcpy(grammar, DEFAULT_CASS_GRAMMAR);
    }
    if (has_suf(grammar, "reg")) load_reg(grammar);
    else load_fsc(grammar);

    if (where) fprintf(stderr, "Grammar file: %s\n", RegFilename);
    HeadFtr = intern_att(Head);


    /**  Infile  **/
    if (infilename != NULL) {
	strcpy(filename, infilename);
	infile = fopen(filename, "r");
	if (infile == NULL) {
	    if (has_suf(filename, "tag")) error("File not found: %s", filename);
	    addsuf(filename, ".tag");
	    infile = fopen(filename, "r");
	    if (infile == NULL) error("File not found: %s", filename);
	}
    }


    /**  Set Output Format  **/
    if (grepcat != NULL) CassGrepCat = intern(grepcat);
    else if (format == NULL || string_equal(format, "forest")) forest_format();
    else if (string_equal(format, "tuples")) tuple_format();
    else if (string_equal(format, "pairs")) pair_format();
    else if (string_equal(format, "roles")) role_format();
    else if (string_equal(format, "args")) arg_format();
    else if (string_equal(format, "vertical")) vertical_format();
    else if (string_equal(format, "binary")) binary_format();
    else if (string_equal(format, "multi")) multi_format();
    else if (string_equal(format, "phrases")) phrases_format();
    else if (string_equal(format, "xkwic")) xkwic_format();
    else error("Unrecognized output format: %s", format);


    /**  Open Outfile(s)  **/
    if (MultiFile)
      init_outfiles(outfilename == NULL ? infilename : outfilename);
    else
      Outfile = outfilename == NULL ? stdout : Fopen(outfilename, "w");


    /**  Do It  **/
    while (scan_sentence(infile) != EOF) {
	match();
	if (print_sentence != NULL) print_sentence();
    }

    exit(0);
}

