#include <stdio.h>
#include "spa.h"
#include "sents.h"


int LeftContext = DEFAULT_LEFT_CONTEXT;
int GutterWidth = DEFAULT_GUTTER_WIDTH;


dekwic (FILE *infile, FILE *outfile)
{
    int c, i, m;
    char cat[64];

    while (peek_char(infile) != EOF) {

	/* Skip to target */
	for (i = 0; i < LeftContext+GutterWidth; i++) {
	    c = scan_char(infile);
	    if (c == EOF || c == '\n') {
		fprintf(stderr, "Warning: line ends prematurely, skipping\n");
		goto nextline;
	    }
	}

	if (isspace(peek_char(infile))) {
	    fprintf(stderr, "Warning: whitespace where target should be, skipping line\n");
	    goto skipline;
	}

	m = 0;
	do {
	    skip_chars_not(" \t\n", infile);
	    skip_chars(" \t", infile);
	    m++;
	}
	while ((c = peek_char(infile)) != ']' && c != EOF && c != '\n');
	if (c == ']') {
	    scan_char(infile); /* ']' */
	    scan_string(cat, 64, " \t\n", infile);
	    fprintf(outfile, "%s %d", cat, m);
	}

      skipline:
	skip_chars_not("\n", infile);
	scan_char(infile); /* '\n' */

      nextline:
	terpri(outfile);
    }
}


char RekwicBracket = ')';

rekwic (FILE *concfile, FILE *markupfile, FILE *outfile)
{
    int m, c, i, k;
    char line[1024];
    struct record r;

    while (scan_record_1(&r, markupfile, " \t", SQUEEZE) != EOF) {
	if (r.nfields == 2) {
	    m = atoi(r.fields[1]);
	    for (i = 0; i < LeftContext+GutterWidth; i++) {
		c = scan_char(concfile);
		if (c == EOF || c == '\n') {
		    error("Premature end of line in concordance file");
		}
		print_char(c, outfile);
	    }
	    for (k = 0; k < m; k++) {
		if (peek_char(concfile) == ']') k--;
		while (!isspace(c = scan_char(concfile)) && c != EOF)
		  print_char(c, outfile);
		if (c != EOF) print_char(c, outfile);
		while (isspace(c = scan_char(concfile)) && c != '\n')
		  print_char(c, outfile);
		if (c != EOF) unread_char(c, concfile);
	    }
	    print_char(')', outfile);
	    print_string(r.fields[0], outfile);
	    print_char(' ', outfile);
	    while ((c = scan_char(concfile)) != EOF && c != '\n')
	      print_char(c, outfile);
	    terpri(outfile);
	}
	else {
	    if (scan_line(line, 1024, concfile) == EOF)
	      error("Premature end of concordance file");
	    if (r.nfields == 1)
	      fprintf(stderr, "Too few fields in markup line, ignoring");
	    fprintf(outfile, "%s\n", line);
	}
    }
}
	    


main (int argc, char *argv[])
{
    char *tmp, *infilename = NULL, *outfilename = NULL, *markup = NULL;
    FILE *infile = stdin, *outfile = stdout, *markupfile;
    
    tmp = getdefault("KwicLeftContext");
    if (tmp != NULL) LeftContext = atoi(tmp);

    tmp = getdefault("KwicGutterWidth");
    if (tmp != NULL) GutterWidth = atoi(tmp);

    parse_args(&argc, &argv, "?s-so-sr-ilc-igw",
	       &infilename, &outfilename, &markup, &LeftContext, &GutterWidth);

    if (infilename != NULL) infile = Fopen(infilename, "r");
    if (outfilename != NULL) outfile = Fopen(outfilename, "w");

    if (markup != NULL) {
	markupfile = Fopen(markup, "r");
	rekwic(infile, markupfile, outfile);
    }
    else dekwic(infile, outfile);

    exit(0);
}
