#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <spastrng.h>  /* string_member */


#define shift (argc--, argv++)

#define read_char getc
#define write_char putc
#define scan_char getc
#define print_char fputc
#define unread_char ungetc

#define string_equal(s1, s2) (strcmp((s1),(s2)) == 0)
#define string_member(x, s) (strchr((s),(x)))
#define print_string fputs


int peek_char (FILE *stream)
{
  int c;
  c = read_char(stream);
  if (c != EOF) unread_char(c, stream);
  return c;
}


int scan_string (char *s, int n, char *delims, FILE *file)
{
    int c, i = 0;

    while ((c = read_char(file)) != EOF && !string_member(c, delims)) {
	if (i >= n) {
	    fprintf(stderr, "Buffer overflow scanning string");
	    exit(1);
	}
	s[i++] = c;
    }
    if (i == 0 && c == EOF) return EOF;
    if (i >= n) {
	fprintf(stderr, "Buffer overflow scanning string");
	exit(1);
    }
    s[i] = '\0';
    return (c == EOF ? 0 : c);
}


char *expand_string (char *s)
{
    static char buf[1024];
    char *v, spc;
    int i, j, o = 0, len;

    for (i = 0; s[i]; i++) {
	if (o >= 1024) {
	    fprintf(stderr, "expand_string: buffer overflow\n");
	    exit(1);
	}

	/** Dollar Sign **/
	if (s[i] == '$') {
	    i++;

	    /** Dollar Escape **/
	    if (s[i] == '$') {
		buf[o++] = '$';
	    }

	    /** Variable Substitution **/
	    else {
		if (s[i] == '{' || s[i] == '(') {
		    i++;
		    for (j = i; s[j] && s[j] != '}' && s[j] != ')'; j++);
		    spc=0;
		}
		else {
		    for (j = i; s[j] && (s[j] == '_' || isalnum(s[j])); j++);
		    spc = s[j];
		}
		s[j] = '\0';
		v = getenv(s+i);
		if (v != NULL) {
		    len = strlen(v);
		    if (o+len >= 1024) {
			fprintf(stderr, "expand_string: buffer overflow\n");
			exit(1);
		    }
		    strcpy(buf+o, v);
		    o += len;
		}
		i = j;
		if (spc) s[i--] = spc;
	    }
	}

	/** Backslash Escape **/
	else if (s[i] == '\\') {
	    i++;
	    if (s[i] == 'n') {
		buf[o++] = '\n';
	    }
	    else if (s[i] == 't') {
		buf[o++] = '\t';
	    }
	    else buf[o++] = s[i];
	}
	    
	else buf[o++] = s[i];
    }
    return buf;
}


main (int argc, char *argv[])
{
    char *inpaths[128], *outpath = NULL, *envpath = NULL;
    FILE *infile = stdin, *outfile = stdout, *file;
    int c, n, i, d = 0;
    char buf[256], *s;

    shift;
    while (argc > 0) {
	if (string_equal(argv[0], "-o")) {
	    shift;
	    outpath = argv[0];
	    shift;
	}
	else if (string_equal(argv[0], "-f")) {
	    shift;
	    envpath = argv[0];
	    shift;
	}
	else if (string_member('=', argv[0])) {
	    putenv(strdup(expand_string(argv[0])));
	    shift;
	}
	else {
	    if (d >= 128) {
		fprintf(stderr, "shpp: too many inpaths");
		exit(1);
	    }
	    inpaths[d++] = argv[0];
	    shift;
	}
    }
	
    if (envpath != NULL) {
	file = fopen(envpath, "r");
	if (file == NULL) {
	    fprintf(stderr, "shpp: can't open envpath %s\n", envpath);
	    exit(1);
	}
	while (scan_string(buf, 256, "\n", file) != EOF) {
	    putenv(strdup(expand_string(buf)));
	}
	fclose(file);
    }

    if (d == 0) inpaths[d++] = "-";

    if (outpath != NULL && !string_equal(outpath, "-")) {
	outfile = fopen(outpath, "w");
	if (outfile == NULL) {
	    fprintf(stderr, "shpp: can't open outpath %s\n", outpath);
	    exit(1);
	}
    }

    while (d > 0) {
	d--;
	if (string_equal(inpaths[d], "-")) {
	    infile = stdin;
	}
	else {
	    infile = fopen(inpaths[d], "r");
	    if (infile == NULL) {
		fprintf(stderr, "shpp: can't open inpath %s\n", inpaths[d]);
		exit(1);
	    }
	}

	while ((c = read_char(infile)) != EOF) {
	    if (c == '$') {
		n = peek_char(infile);
		if (n == '$') {
		    read_char(infile);
		    write_char('$', outfile);
		}
		else if (n == '{' || n == '(') {
		    read_char(infile);
		    scan_string(buf, 256, "}) \t\n", infile);
		    s = getenv(buf);
		    if (s != NULL) print_string(s, outfile);
		}
		else {
		    n = 0;
		    while (c = read_char(infile), isalnum(c) || c == '_') {
			buf[n++] = c;
		    }
		    buf[n] = '\0';
		    if (c != EOF) unread_char(c, infile);
		    s = getenv(buf);
		    if (s != NULL) print_string(s, outfile);
		}
	    }
	    else write_char(c, outfile);
	}

	if (infile != stdin) fclose(infile);
    }
    
    if (outfile != stdout) fclose(outfile);
    exit(0);
}
