#include <spa.h>
#include <smlsyntax.h>


main (int argc, char *argv[])
{
    struct lined_file *file;
    
    parse_args(&argc, &argv, "*");

    if (argc == 0) smllint(lopen(NULL));
    else {
	for (; argc > 0; SHIFT) {
	    file = lopen(argv[0]);
	    smllint(file);
	    lclose(file);
	}
    }
}

    


#define MAXNAME 64
#define MAXDEPTH 128


int scan_word (char *buf, struct lined_file *file)
{
    int c;
    int first = 1, i = 0, quoted = 0;

    if (string_member(lpeek_char(file), Prenectors)) {
	buf[i++] = lread_char(file);
    }

    c = lpeek_char(file);
    if (!isalnum(c) && !string_member(c, Connectors)) {
	buf[i++] = lread_char(file);
    }
    else {
	while ((c = lread_char(file)) != EOF) {
	    if (!isalnum(c) && !string_member(c, Connectors)) {
		if (c != Terminator) lunread_char(c, file);
		break;
	    }
	    if (i >= MAXNAME-1) error("smllint: MAXNAME exceeded");
	    buf[i++] = c;
	}
    }

    buf[i] = '\0';
    if (i == 0) printf("%s, line %d: empty word\n", file->filename, file->line);
}



struct entry {
    char name[MAXNAME];
    int start;
};


smllint (struct lined_file *file)
{
    int c, top = -1;
    struct entry stack[MAXDEPTH];
    char tmp[MAXNAME];
    
    while ((c = lread_char(file)) != EOF) {
	if (c == '^') {
	    c = lread_char(file);
	    if (c == '[') {
		top++;
		stack[top].start = file->line;
		scan_word(stack[top].name, file);
	    }
	    else if (c == ']') {
		scan_word(tmp, file);
		if (top < 0) printf("%s, line %d: unmatched ]%s\n",
				    file->filename, file->line, tmp);
		else if (string_equal(tmp, stack[top].name)) top--;
		else if (top > 0 && string_equal(tmp, stack[top-1].name)) {
		    printf("%s, line %d: apparently unmatched [%s, detected at ]%s, line %d\n",
			   file->filename, stack[top].start, stack[top].name, tmp, file->line);
		    top -= 2;
		}
		else printf("%s, line %d: apparently unmatched ]%s\n",
			    file->filename, file->line, tmp);
	    }
	}
    }
    while (top > -1) {
	printf("%s, line %d: unmatched [%s, detected at EOF\n",
	       file->filename, stack[top].start, stack[top].name);
	top--;
    }
}

		
		    
