#include <stdio.h>
#include <ctype.h>
#include <spa.h>


#define BUFFER_LEN 256

int Trivial = 0;
int Cased = 0;
int NoNum = 0;
int NoSpace = 0;


/**================================================================**/
/**  Fetch and Lookahead                                           **/
/**                                                                **/

FILE *Infile = NULL;
char Buffer[BUFFER_LEN] = { '\0' };
int B = 0;
int La = 0;

int peek ()
{
    int x = fetch();
    if (x != EOF) unfetch(1);
    return x;
}

int fetch ()
{
    int c = Buffer[B];

    if (c == '\0') {
	c = read_char(Infile);
	if (c == EOF) return EOF;
	Buffer[B] = c;
	B = (B+1)%BUFFER_LEN;
	Buffer[B] = '\0';
    }
    else B = (B+1)%BUFFER_LEN;
    return c;
}

int unfetch (int n)
{
    int i;

    for (i = 0; i < n; i++) {
	if (--B < 0) B += BUFFER_LEN;
	if (Buffer[B] == '\0') error("buffer overflow");
    }
}

int lookahead ()
{
    int x = fetch();
    if (x != EOF) La++;
    return x;
}

#define ALL -1
int unlookahead (int n)
{
    if (n == ALL) n = La;
    if (n > La) error("unlookahead more than lookahead");
    unfetch(n);
    La -= n;
}

#define keep_lookahead() (La = 0)
    

/**================================================================**/
/**  Strings                                                       **/
/**                                                                **/

print_lower (char *s)
{
    while (*s) {
	print_char(isupper(*s) ? tolower(*s++) : *s++, stdout);
    }
}

print_visibly (char *s)
{
    for (; *s; s++) {
	if (isprint(*s)) {
	    switch(*s) {
	      case '\t':
		printf("\\t");
		break;
	      case ' ':
		printf("\\s");
		break;
	      case '\\':
		printf("\\\\");
		break;
	      default:
		print_char(*s, stdout);
	    }
	}
	else if (*s == '\n') {
	    printf("\\n");
	}
	else printf("\\%o", *s);
    }
}


/**================================================================**/
/**  whitespace containing multiple newlines: \VS text             **/
/**  whitespace containing newline followed by space/tab:          **/
/**      \indent text                                              **/
/**  multiple -'s: \dash text                                      **/
/**  - preceded and followed by alphanumeric, possibly with        **/
/**      intervening \n following: \hyphen text                    **/
/**  - otherwise: \dash text                                       **/
/**  (+-)? numeric+ (, numeric^3)* (. numeric+)?: \num text        **/
/**  (+-)? . numeric+: \num text                                   **/
/**  ' surrounded by alphanumeric: in word                         **/
/**  ``: token                                                     **/
/**  '': token                                                     **/
/**  \: \backslash                                                 **/
/**  regular token: text text caps following-space                 **/


int BOF = 1;
int AfterWhitespace = 1;

tokenize ()
{
    whitespace();

    while (1) {
	if (number()) ;
	else if (hyphen()) ;
	else if (word()) ;
	else if (sgml()) ;
	else if (punc()) ;
	else break;
	whitespace();
    }
}


whitespace ()
{
    char buf[BUFFER_LEN];
    int i = 0, newlines = 0, printed = 0, x;

    while (isspace(x = fetch())) {
	buf[i++] = x;
	if (x == '\n') newlines++;
    }
    if (x != EOF) unfetch(1);
    buf[i] = '\0';

    if (Trivial) {
	if (!BOF) terpri(stdout);
    }
    else {
	if (!NoSpace) {
	    if (newlines > 1 || BOF && newlines > 0) {
		if (!BOF) terpri(stdout);
		printf("\\VS\t");
		print_visibly(buf);
		terpri(stdout);
		printed = 1;
	    }
	    if ((newlines > 0 || BOF && i > 0)
		&& (buf[i-1] == ' ' || buf[i-1] == '\t')) {
		if (!BOF) terpri(stdout);
		printf("\\indent\t");
		print_visibly(buf);
		terpri(stdout);
		printed = 1;
	    }
	}
	if (!printed && !BOF) {
	    print_char('\t', stdout);
	    if (i == 0) print_char('-', stdout);
	    else print_visibly(buf);
	    terpri(stdout);
	    printed = 1;
	}
    }
    AfterWhitespace = i;
    BOF = 0;
}

int number ()
{
    int x, n = 0, comma = 0, sincecomma = 0, point = 0;
    char plusminus = 0;
    char buf[BUFFER_LEN];

    if ((x = lookahead()) == '+' || x == '-') plusminus = x, x = lookahead();
    if (isdigit(x)) {
	unlookahead(1);
	while (1) {
	    buf[n++] = x = lookahead();
	    if (isdigit(x)) {
		if (comma) {
		    sincecomma++;
		    if (sincecomma > 3) {
			unlookahead(sincecomma+1);
			n -= sincecomma+1;
			goto WrapUp;
		    }
		}
	    }
	    else if (x == ',') {
		if (comma) {
		    if (sincecomma != 3) {
			unlookahead(sincecomma+2);
			n -= sincecomma+2;
			goto WrapUp;
		    }
		    else sincecomma = 0;
		}
		else comma = 1,	sincecomma = 0;
	    }
	    else {
		n--;
		break;
	    }
	}
	if (comma && sincecomma != 3) {
	    unlookahead(sincecomma+1);
	    n -= sincecomma+1;
	    goto WrapUp;
	}
    }
    if (x == '.') {
	point = 1;
	buf[n++] = x;
	while (isdigit(x = lookahead())) buf[n++] = x;
    }
    if (isalpha(x) && !plusminus && !point) {
	unlookahead(ALL);
	return 0;
    }
    else if (x != EOF) unlookahead(1);

  WrapUp:
    buf[n] = '\0';
    if (n == 0 || n == 1 && point) {
	unlookahead(ALL);
	return 0;
    }
    else {
	if (NoNum) {
	    if (plusminus) print_char(plusminus, stdout);
	    print_string(buf, stdout);
	}
	else printf("\\num");
	if (!Trivial) {
	    print_char('\t', stdout);
	    if (plusminus) print_char(plusminus, stdout);
	    print_string(buf, stdout);
	}
	keep_lookahead();
	return 1;
    }
}
    
int hyphen ()
{
    int x, i, n = 0;

    if ((x = fetch()) != '-') {
	if (x != EOF) unfetch(1);
	return 0;
    }
    n = 1;
    while ((x = fetch()) == '-') n++;
    if (x != EOF) unfetch(1);

    if (n > 1) {
	if (!Trivial) printf("\\dash\t");
	for (i = 0; i < n; i++) print_char('-', stdout);
    }

    else {
	int hyphen = 0;

	if (!AfterWhitespace) {
	    while ((x = lookahead()) == ' ');
	    if (x == '\n') {
		x = lookahead();
		if (isalnum(x)) hyphen = 1;
	    }
	    unlookahead(ALL);
	}
	if (!Trivial) {
	    if (hyphen) printf("\\hyphen\t");
	    else printf("\\dash\t");
	}
	print_char('-', stdout);
    }

    return 1;
}

int word ()
{
    int x, n = 0;
    char buf[BUFFER_LEN];

    while (isalnum(x = fetch()) || x == '\'' && n > 0) {
	buf[n++] = x;
    }
    if (x != EOF) unfetch(1);
    while (n > 0 && buf[n-1] == '\'') unfetch(1), n--;
    if (n == 0) return 0;
    buf[n] = '\0';
    if (Cased) print_string(buf, stdout);
    else print_lower(buf);
    if (!Trivial) {
	print_char('\t', stdout);
	print_string(buf, stdout);
    }
    return 1;
}

int punc ()
{
    int x;
    char buf[2], *s;

    x = fetch();
    if (x == EOF) return 0;
    if (isspace(x)) {
	unfetch(1);
	return 0;
    }
    if (x == '`') {
	if (peek() == '`') {
	    fetch();
	    printf("``");
	    if (!Trivial) printf("\t``");
	    return 1;
	}
    }
    else if (x == '\'') {
	if (peek() == '\'') {
	    fetch();
	    printf("'");
	    if (!Trivial) printf("\t'");
	    return 1;
	}
    }
    buf[0] = x;
    buf[1] = '\0';
    print_visibly(buf);
    if (!Trivial) {
	print_char('\t', stdout);
	print_visibly(buf);
    }
    return 1;
}

sgml ()
{
    char buf[1024];
    int x, n = 0;

    x = fetch();
    if (x != '<') {
	if (x != EOF) unfetch(1);
	return 0;
    }
    buf[n++] = x;
    while ((x = fetch()) != '>' && x != '\n' && x != EOF) {
	buf[n++] = x;
    }
    if (x == '>') buf[n++] = x;
    else if (x != EOF) unfetch(1);
    buf[n] = '\0';
    print_visibly(buf);
    if (!Trivial) {
	print_char('\t', stdout);
	print_visibly(buf);
    }
    return 1;
}


/**================================================================**/
/**  Main                                                          **/
/**                                                                **/

main (int argc, char *argv[])
{
    char *filename = NULL;

    Usage = "token [options] [filename]\n\
\n\
Tokenizes the named file, or else stdin.\n\
\n\
In the basic format, the first field is the word or the name of\n\
a special token (\\VS, \\indent, \\num).  Words are downcased.\n\
The second field is the original string for the token, and the\n\
third field contains following whitespace printed readably.\n\
SGML directives are passed through.\n\
\n\
-c means preserve case in the first column.\n\
-n means preserve numbers in the first column.\n\
-s means suppress VS and indent.\n\
-0 means do all three and print only the second column.";

    parse_args(&argc, &argv, "?s-bc-bn-bs-b0",
	       &filename, &Cased, &NoNum, &NoSpace, &Trivial);

    if (Trivial) Cased = NoNum = NoSpace = 1;
    Infile = filename == NULL ? stdin : Fopen(filename, "r");
    tokenize();
}
