#include <spa.h>
#include <cass.h>


main (int argc, char *argv[])
{
    char *infilename = NULL, *outfilename = NULL, *grammar, *fxfile = NULL;
    char *format = NULL, filename[MAXFILENAME];
    int fx = 0, vert = 0, phrases = 0, nokids = 0, where = 0;
    FILE *infile = stdin;

    grammar = "cass.reg";
    DontPrintRoles = 0;
    DontPrintAttributes = 0;

    Usage = "cass [infile] [-g <grammar>] [-w] [-o <outfile>]\n\
    [-rsvpn] [-format <format>] [-ch]\n\
infile defaults to stdin\n\
-g default is value of 'CassGrammar'\n\
-w 'where' shows filename of grammar being used\n\
-o defaults to stdout\n\
-r means recognize only, don't parse\n\
-s means silent (no warnings)\n\
-format legal values are: forest [default], vertical, multi, binary, xkwic, phrases\n\
-v is shorthand for '-format vertical'\n\
-p is shorthand for '-format phrases'\n\
-ch suppresses children in phrases format\n\
-n suppresses printing of corpus positions";

    parse_args(&argc, &argv, "?s-sg-sf-so-bdebug-br-bs-sformat-bv-bp-bch-sh-bw-bn",
	       &infilename, &grammar, &grammar, &outfilename, &MatchVerbose, &NoActions,
	       &NoWarn, &format, &vert, &phrases, &nokids, &Head, &where, &DontPrintCPs);

#ifndef DEBUG
    if (MatchVerbose) printf("Sorry, not compiled DEBUG, no verbose");
#endif
    if (phrases) format = "phrases";
    if (vert) format = "vertical";
    if (nokids) PrintChildren = 0;


    /**  Grammar  **/
    if (grammar == NULL) error("%s: no grammar specified, no default", ProgName);
    if (has_suf(grammar, "reg")) load_reg(grammar);
    else load_fsc(grammar);

    if (where) fprintf(stderr, "Grammar file: %s\n", RegFilename);
    HeadFtr = intern_att(Head);


    /**  Infile  **/
    if (infilename != NULL) {
	strcpy(filename, infilename);
	infile = fopen(filename, "r");
	if (infile == NULL) {
	    if (has_suf(filename, "tag")) error("File not found: %s", filename);
	    addsuf(filename, ".tag");
	    infile = fopen(filename, "r");
	    if (infile == NULL) error("File not found: %s", filename);
	}
    }


    /**  Set Output Format  **/
    if (format == NULL || string_equal(format, "forest")) forest_format();
    else if (string_equal(format, "tuples")) tuple_format();
    else if (string_equal(format, "pairs")) pair_format();
    else if (string_equal(format, "roles")) role_format();
    else if (string_equal(format, "args")) arg_format();
    else if (string_equal(format, "vertical")) vertical_format();
    else if (string_equal(format, "binary")) binary_format();
    else if (string_equal(format, "multi")) multi_format();
    else if (string_equal(format, "phrases")) phrases_format();
    else if (string_equal(format, "xkwic")) xkwic_format();
    else error("Unrecognized output format: %s", format);


    /**  Open Outfile(s)  **/
    if (MultiFile)
      init_outfiles(outfilename == NULL ? infilename : outfilename);
    else
      Outfile = outfilename == NULL ? stdout : Fopen(outfilename, "w");


    /**  Do It  **/
    while (scan_sentence(infile) != EOF) {
	match();
	if (print_sentence != NULL) print_sentence();
    }

    exit(0);
}

